# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""The lxp module contains classes and functions to simplify the interaction with ray data.

Ray data is provided as lpf file.
LPF files contain a set of simulated rays with all their intersections and properties.
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union

import ansys.api.speos.lpf.v2.lpf_file_reader_pb2 as lpf_file_reader__v2__pb2
import ansys.api.speos.lpf.v2.lpf_file_reader_pb2_grpc as lpf_file_reader__v2__pb2_grpc
from ansys.speos.core.generic.general_methods import graphics_required
from ansys.speos.core.project import Project, Speos

if TYPE_CHECKING:  # pragma: no cover
    from ansys.tools.visualization_interface import Plotter
try:
    from ansys.speos.core.generic.general_methods import run_if_graphics_required

    run_if_graphics_required(warning=True)
except ImportError as err:  # pragma: no cover
    raise err

ERROR_IDS = [7, 8, 9, 10, 11, 12, 13, 14, 15]
"""Intersection types indicating an error state."""

NO_ERROR_IDS = [0, 1, 2, 3, 4, 5, 6, 16, -7, -6, -5, -5, -4, -3, -2, -1]
"""Intersection types indicating a correct ray state."""


class RayPath:
    """Framework representing a singular ray path.

    Parameters
    ----------
    raypath : ansys.api.speos.lpf.v2.lpf_file_reader__v2__pb2.RayPath
        RayPath object
    sensor_contribution : bool
        Defines if sensor contributions are stored within the data.
        By default ``False``.
    """

    def __init__(
        self,
        raypath: lpf_file_reader__v2__pb2.RayPath,
        sensor_contribution: bool = False,
    ):
        self._nb_impacts = len(raypath.impacts)
        self._impacts = [[inter.x, inter.y, inter.z] for inter in raypath.impacts]
        self._wl = raypath.wavelengths[0]
        self._body_ids = raypath.body_context_ids
        self._face_ids = raypath.unique_face_ids
        self._last_direction = [
            raypath.lastDirection.x,
            raypath.lastDirection.y,
            raypath.lastDirection.z,
        ]
        self._intersection_type = raypath.interaction_statuses
        if sensor_contribution:
            self._sensor_contribution = [
                {
                    "sensor_id": sc.sensor_id,
                    "position": [sc.coordinates.x, sc.coordinates.y],
                }
                for sc in raypath.sensor_contributions
            ]
        else:
            self._sensor_contribution = None

    @property
    def nb_impacts(self) -> int:
        """Number of impacts contained in ray path.

        Returns
        -------
        int
            Number of impacts
        """
        return self._nb_impacts

    @property
    def impacts(self) -> list[list[float]]:
        """XYZ coordinates for each impact.

        Returns
        -------
        list[list[float]]
            list containing the impact coordinates [[x0,y0,z0],[x1,y1,z1],...]
        """
        return self._impacts

    @property
    def wl(self) -> float:
        """Wavelength of the ray.

        Returns
        -------
        float
            Wavelength in nm
        """
        return self._wl

    @property
    def body_ids(self) -> list[int]:
        """Body ID of interacted body for each impact.

        Returns
        -------
        list[int]
            List of body IDs for each impact.
        """
        return self._body_ids

    @property
    def face_ids(self) -> list[int]:
        """Face ID of interacted body for each impact.

        Returns
        -------
        list[int]
            List of face IDs for each impact.
        """
        return self._face_ids

    @property
    def last_direction(self) -> list[float]:
        """Last direction of the ray.

        Returns
        -------
        list[float]
            Last direction of the rays as list[x,y,z].
        """
        return self._last_direction

    @property
    def intersection_type(self) -> list[int]:
        """Intersection type of the ray for each impact.

        Returns
        -------
        list[int]
            Intersection type at each impact.

        Notes
        -----
        Available intersection types:

        - StatusAbsorbed = 0
        - StatusSpecularTransmitted = 1
        - StatusGaussianTransmitted = 2
        - StatusLambertianTransmitted = 3
        - StatusVolumicDiffused = 4
        - StatusJustEmitted = 5
        - StatusDiracTransmitted = 6
        - StatusError = 7
        - StatusErrorVolumicBodyNotClosed = 8
        - StatusErrorVolumeConflict = 9
        - StatusError2DTangency = 10
        - StatusError2DIntersect3DWarning = 11
        - StatusErrorNonOpticalMaterial = 12
        - StatusErrorIntersection = 13
        - StatusErrorNonOpticalMaterialAtEmission = 14
        - StatusError3DTextureSupportTangency = 15
        - StatusLast = 16
        - StatusFirst = -7
        - StatusDiracReflected = -6
        - StatusReserved = -5
        - StatusGrinStep = -4
        - StatusLambertianReflected = -3
        - StatusGaussianReflected = -2
        - StatusSpecularReflected = -1
        """
        return self._intersection_type

    @property
    def sensor_contribution(self) -> Union[None, list[dict]]:
        """Provide the sensor contribution information for each sensor.

        Returns
        -------
        Union[None, list[dict]]
            If no sensor contribution, None will be returned. If there is sensor contribution, \
            a dictionary with the following information is returned:\
            {“sensor_id”: sc.sensor_id,
            “position”: [sc.coordinates.x, sc.coordinates.y]}
        """
        return self._sensor_contribution

    def get(self, key=""):
        """Retrieve any information from the RayPath object.

        Parameters
        ----------
        key : str
            Name of the property.

        Returns
        -------
        property
            Values/content of the associated property.
        """
        data = {k: v.fget(self) for k, v in RayPath.__dict__.items() if isinstance(v, property)}
        if key == "":
            return data
        elif data.get(key):
            return data.get(key)
        else:
            print("Used key: {} not found in key list: {}.".format(key, data.keys()))

    def __str__(self):
        """Create string representation of a RayPath."""
        return str(self.get())


class LightPathFinder:
    """Define an interface to read LPF files.

    LPF files contain a set of simulated rays including their intersections and properties.

    Parameters
    ----------
    speos : ansys.speos.core.speos.Speos
        Speos Session (connected to Speos gRPC server).
    path : str
        Path to the LPF file to be opened.

    """

    def __init__(self, speos: Speos, path: str):
        self.client = speos.client
        """Speos instance client"""
        self._stub = lpf_file_reader__v2__pb2_grpc.LpfFileReader_MonoStub(self.client.channel)
        self.__open(path)
        self._data = self._stub.GetInformation(
            lpf_file_reader__v2__pb2.GetInformation_Request_Mono()
        )
        self._nb_traces = self._data.nb_of_traces
        self._nb_xmps = self._data.nb_of_xmps
        self._has_sensor_contributions = self._data.has_sensor_contributions
        self._sensor_names = self._data.sensor_names
        self._rays = self.__parse_traces()
        self._filtered_rays = []

    @property
    def nb_traces(self) -> int:
        """Number of light paths within LPF data set."""
        return self._nb_traces

    @property
    def nb_xmps(self) -> int:
        """Number of sensors involved within LPF data set."""
        return self._nb_xmps

    @property
    def has_sensor_contributions(self) -> bool:
        """Define if a LPF file contains information regarding the sensor contribution."""
        return self._has_sensor_contributions

    @property
    def sensor_names(self) -> list[str]:
        """List of involved sensor names."""
        return self._sensor_names

    @property
    def rays(self) -> list[RayPath]:
        """List ray paths within LPF file."""
        return self._rays

    @property
    def filtered_rays(self) -> list[RayPath]:
        """List of filtered ray paths."""
        return self._filtered_rays

    def __str__(self):
        """Create string representation of LightPathFinder."""
        return str(
            {
                k: v.fget(self)
                for k, v in LightPathFinder.__dict__.items()
                if isinstance(v, property) and "rays" not in k
            }
        )

    def __open(self, path: str):
        """Open LPF file.

        Parameters
        ----------
        path : str
            Path to file
        """
        self._stub.InitLpfFileName(
            lpf_file_reader__v2__pb2.InitLpfFileName_Request_Mono(lpf_file_uri=path)
        )

    def __parse_traces(self) -> list[RayPath]:
        """Read all ray paths from lpf dataset.

        Returns
        -------
            list[script.RayPath]

        """
        raypaths = []
        for rp in self._stub.Read(lpf_file_reader__v2__pb2.Read_Request_Mono()):
            raypaths.append(RayPath(rp, self._has_sensor_contributions))
        return raypaths

    def __filter_by_last_intersection_types(self, options: list[int], new=True):
        """Filter ray paths based on last intersection types.

        Populate filtered_rays property.
        """
        if new:
            self._filtered_rays = []
            for ray in self._rays:
                if int(ray.intersection_type[-1]) in options:
                    self._filtered_rays.append(ray)
        else:
            temp_rays = self._filtered_rays
            self._filtered_rays = []
            for ray in temp_rays:
                if int(ray.intersection_type[-1]) in options:
                    self._filtered_rays.append(ray)

    def filter_by_face_ids(self, options: list[int], new=True) -> LightPathFinder:
        """Filter ray paths based on face IDs and populates filtered_rays property.

        Parameters
        ----------
        options : list[int]
            List of face IDs.
        new : bool
            Define if a new filter is created or an existing filter is filtered.

        Returns
        -------
        ansys.speos.core.lxp.LightPathFinder
            LightPathFinder Instance.
        """
        if new:
            self._filtered_rays = []
            for ray in self._rays:
                if any([faceid in options for faceid in ray.face_ids]):
                    self._filtered_rays.append(ray)
        else:
            temp_rays = self._filtered_rays
            self._filtered_rays = []
            for ray in temp_rays:
                if any([faceid in options for faceid in ray.face_ids]):
                    self._filtered_rays.append(ray)
        return self

    def filter_by_body_ids(self, options: list[int], new=True) -> LightPathFinder:
        """Filter ray paths based on body IDs and populates filtered_rays property.

        Parameters
        ----------
        options : list[int]
            List of body IDs.
        new : bool
            Define if a new filter is created or an existing filter is filtered.

        Returns
        -------
        ansys.speos.core.lxp.LightPathFinder
            LightPathFinder Instance.
        """
        if new:
            self._filtered_rays = []
            for ray in self._rays:
                if any([body_id in options for body_id in ray.body_ids]):
                    self._filtered_rays.append(ray)
        else:
            temp_rays = self._filtered_rays
            self._filtered_rays = []
            for ray in temp_rays:
                if any([body_id in options for body_id in ray.body_ids]):
                    self._filtered_rays.append(ray)
        return self

    def filter_error_rays(self) -> LightPathFinder:
        """Filter ray paths and only shows rays in error.

        Returns
        -------
        ansys.speos.core.lxp.LightPathFinder
            LightPathFinder Instance.
        """
        self.__filter_by_last_intersection_types(options=ERROR_IDS)
        return self

    def remove_error_rays(self) -> LightPathFinder:
        """Filter rays and only shows rays not in error.

        Returns
        -------
        ansys.speos.core.lxp.LightPathFinder
            LightPathFinder Instance.
        """
        self.__filter_by_last_intersection_types(options=NO_ERROR_IDS)
        return self

    @staticmethod
    @graphics_required
    def __add_ray_to_pv(plotter: Plotter, ray: RayPath, max_ray_length: float):
        """Add a ray to pyvista plotter.

        Parameters
        ----------
        plotter : Plotter
            Ansys plotter object to which rays should be added.
        ray : script.RayPath
            RayPath object which contains ray information to be added.
        max_ray_length : float
            Length of the last ray.
        """
        import pyvista as pv

        temp = ray.impacts.copy()
        if not 7 <= ray.intersection_type[-1] <= 15:
            temp.append(
                [
                    ray.impacts[-1][0] + max_ray_length * ray.last_direction[0],
                    ray.impacts[-1][1] + max_ray_length * ray.last_direction[1],
                    ray.impacts[-1][2] + max_ray_length * ray.last_direction[2],
                ]
            )
        if len(ray.impacts) > 2:
            mesh = pv.MultipleLines(temp)
        else:
            mesh = pv.Line(temp[0], temp[1])
        plotter.plot(mesh, color=wavelength_to_rgb(ray.wl), line_width=2)

    @graphics_required
    def preview(
        self,
        nb_ray: int = 100,
        max_ray_length: float = 50.0,
        ray_filter: bool = False,
        project: Project = None,
        screenshot: Optional[Union[str, Path]] = None,
    ) -> LightPathFinder:
        """Preview LPF file with pyvista.

        Parameters
        ----------
        nb_ray : int
            Number of rays to be visualized.
        max_ray_length : float
            Length of last ray.
        ray_filter : bool
            Boolean to decide if filtered rays or all rays should be shown.
        project : ansys.speos.core.project.Project
            Speos Project/Geometry to be added to pyvista visualisation.
        screenshot : str or Path or ``None``
            Path to save a screenshot of the plotter. If defined Plotter will only create the
            screenshot

        Returns
        -------
        ansys.speos.core.lxp.LightPathFinder
            LightPathFinder Instance.

        Notes
        -----
        Please use the ``q``-key to close the plotter as some
        operating systems (namely Windows) will experience issues
        saving a screenshot if the exit button in the GUI is pressed.
        """
        from ansys.tools.visualization_interface import Plotter

        if ray_filter:
            if len(self._filtered_rays) > 0:
                temp_rays = self._filtered_rays
            else:
                print("no filtered rays")
                temp_rays = self._rays
        else:
            temp_rays = self._rays
        if not project:
            plotter = Plotter()
            if nb_ray > len(temp_rays):
                for ray in temp_rays:
                    self.__add_ray_to_pv(plotter, ray, max_ray_length)
            else:
                for i in range(nb_ray):
                    self.__add_ray_to_pv(plotter, temp_rays[i], max_ray_length)
        else:
            plotter = project._create_preview(viz_args={"opacity": 0.5})
            if nb_ray > len(temp_rays):
                for ray in temp_rays:
                    self.__add_ray_to_pv(plotter, ray, max_ray_length)
            else:
                for i in range(nb_ray):
                    self.__add_ray_to_pv(plotter, temp_rays[i], max_ray_length)
        plotter.show(screenshot=screenshot)
        return self


def wavelength_to_rgb(wavelength: float, gamma: float = 0.8) -> [int, int, int, int]:
    """Convert a given wavelength of light to an approximate RGB color value.

    The wavelength must be given in nanometers in the range from 380 nm to 750 nm.
    Based on the code from http://www.physics.sfasu.edu/astro/color/spectra.html

    Parameters
    ----------
    wavelength : float
        Wavelength in nanometer between 380-750 nm
    gamma : float
        Gamma value.
        By default : ``0.8``
    """
    wavelength = float(wavelength)
    if 380 <= wavelength <= 440:
        attenuation = 0.3 + 0.7 * (wavelength - 380) / (440 - 380)
        r = ((-(wavelength - 440) / (440 - 380)) * attenuation) ** gamma
        g = 0.0
        b = (1.0 * attenuation) ** gamma
    elif 440 <= wavelength <= 490:
        r = 0.0
        g = ((wavelength - 440) / (490 - 440)) ** gamma
        b = 1.0
    elif 490 <= wavelength <= 510:
        r = 0.0
        g = 1.0
        b = (-(wavelength - 510) / (510 - 490)) ** gamma
    elif 510 <= wavelength <= 580:
        r = ((wavelength - 510) / (580 - 510)) ** gamma
        g = 1.0
        b = 0.0
    elif 580 <= wavelength <= 645:
        r = 1.0
        g = (-(wavelength - 645) / (645 - 580)) ** gamma
        b = 0.0
    elif 645 <= wavelength <= 750:
        attenuation = 0.3 + 0.7 * (750 - wavelength) / (750 - 645)
        r = (1.0 * attenuation) ** gamma
        g = 0.0
        b = 0.0
    else:
        r = 0.0
        g = 0.0
        b = 0.0
    r *= 255
    g *= 255
    b *= 255
    return [int(r), int(g), int(b), 255]



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""General methods and helpers collection.

this includes decorator and methods
"""

from collections.abc import Collection
from functools import lru_cache, wraps
import os
from pathlib import Path
from typing import List, Optional, Union, cast
import warnings

import numpy as np

from ansys.speos.core.generic.constants import DEFAULT_VERSION
from ansys.tools.path import get_available_ansys_installations

_GRAPHICS_AVAILABLE = None

GRAPHICS_ERROR = (
    "Preview unsupported without 'ansys-tools-visualization_interface' installed. "
    "You can install this using `pip install ansys-speos-core[graphics]`."
)


def deprecate_kwargs(old_arguments: dict, removed_version="0.3.0"):
    """Issues deprecation warnings for arguments.

    Parameters
    ----------
    old_arguments : dict
        key old argument value new argument name
    removed_version : str
        Release version with which argument support will be removed
        By Default, next major release

    """

    def decorator(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            func_name = function.__name__
            for alias, new in old_arguments.items():
                if alias in kwargs:
                    if new in kwargs:
                        msg = f"{func_name} received both {alias} and {new} as arguments!\n"
                        msg += f"{alias} is deprecated, use {new} instead."
                        raise TypeError(msg)
                    msg = f"Argument `{alias}` is deprecated for method `{func_name}`; it will be "
                    msg += f"removed with Release v{removed_version}. Please use `{new}` instead."
                    kwargs[new] = kwargs.pop(alias)
                    warnings.warn(msg, DeprecationWarning, stacklevel=2)
            retval = function(*args, **kwargs)
            return retval

        return wrapper

    return decorator


@lru_cache
def run_if_graphics_required(warning=False):
    """Check if graphics are available."""
    global _GRAPHICS_AVAILABLE
    if _GRAPHICS_AVAILABLE is None:
        try:
            import pyvista as pv  # noqa: F401

            from ansys.tools.visualization_interface import Plotter  # noqa: F401

            _GRAPHICS_AVAILABLE = True
        except ImportError:  # pragma: no cover
            _GRAPHICS_AVAILABLE = False

    if _GRAPHICS_AVAILABLE is False and warning is False:  # pragma: no cover
        raise ImportError(GRAPHICS_ERROR)
    elif _GRAPHICS_AVAILABLE is False:  # pragma: no cover
        warnings.warn(GRAPHICS_ERROR)


def graphics_required(method):
    """Decorate a method as requiring graphics.

    Parameters
    ----------
    method : callable
        Method to decorate.

    Returns
    -------
    callable
        Decorated method.
    """

    def wrapper(*args, **kwargs):
        run_if_graphics_required()
        return method(*args, **kwargs)

    return wrapper


def magnitude_vector(vector: Collection[float]) -> float:
    """Compute the magnitude (length) of a 2D or 3D vector using NumPy.

    Parameters
    ----------
    vector: List[float]
        A 2D or 3D vector as a list [x, y] or [x, y, z].

    Returns
    -------
    float
        The magnitude (length) of the vector.
    """
    vector_np = np.array(vector, dtype=float)
    if vector_np.size not in (2, 3):
        raise ValueError("Input vector must be either 2D or 3D")
    return float(np.linalg.norm(vector_np))


def normalize_vector(vector: Collection[float]) -> List[float]:
    """
    Normalize a 2D or 3D vector to have a length of 1 using NumPy.

    Parameters
    ----------
    vector: List[float]
        A vector as a list [x, y] for 2D or [x, y, z] for 3D.

    Returns
    -------
    List[float]
        The normalized vector.
    """
    vector_np = np.array(vector, dtype=float)
    if vector_np.size not in (2, 3):
        raise ValueError("Input vector must be either 2D or 3D")

    magnitude = magnitude_vector(vector_np)
    if magnitude == 0:
        raise ValueError("Cannot normalize the zero vector")

    return cast(List[float], (vector_np / magnitude).tolist())


def error_no_install(install_path: Union[Path, str], version: Union[int, str]):
    """Raise error that installation was not found at a location.

    Parameters
    ----------
    install_path : Union[Path, str]
        Installation Path
    version : Union[int, str]
        Version
    """
    install_loc_msg = ""
    if install_path:
        install_loc_msg = f"at {Path(install_path).parent}"
    raise FileNotFoundError(
        f"Ansys Speos RPC server installation not found{install_loc_msg}. "
        f"Please define AWP_ROOT{version} environment variable"
    )


def retrieve_speos_install_dir(
    speos_rpc_path: Optional[Union[Path, str]] = None, version: str = DEFAULT_VERSION
) -> Path:
    """Retrieve Speos install location based on Path or Environment.

    Parameters
    ----------
    speos_rpc_path : Optional[str, Path]
        location of Speos rpc executable
    version : Union[str, int]
        The Speos server version to run, in the 3 digits format, such as "242".
        If unspecified, the version will be chosen as
        ``ansys.speos.core.kernel.client.LATEST_VERSION``.

    """
    if not speos_rpc_path:
        speos_rpc_path = ""

    if not speos_rpc_path or not Path(speos_rpc_path).exists():
        if not Path(speos_rpc_path).exists():
            warnings.warn(
                "Provided executable location not found, looking for local installation",
                UserWarning,
            )
        versions = get_available_ansys_installations()
        ansys_loc = versions.get(int(version), "")
        if not ansys_loc:
            ansys_loc = os.environ.get("AWP_ROOT{}".format(version), "")
            if not ansys_loc:
                error_no_install(speos_rpc_path, int(version))

        speos_rpc_path = Path(ansys_loc) / "Optical Products" / "SPEOS_RPC"
    elif Path(speos_rpc_path).is_file():
        if "SpeosRPC_Server" not in Path(speos_rpc_path).name:
            error_no_install(speos_rpc_path, int(version))
        else:
            speos_rpc_path = Path(speos_rpc_path).parent

    speos_rpc_path = Path(speos_rpc_path)
    if os.name == "nt":
        speos_exec = speos_rpc_path / "SpeosRPC_Server.exe"
    else:
        speos_exec = speos_rpc_path / "SpeosRPC_Server.x"
    if not speos_exec.is_file():
        error_no_install(speos_rpc_path, int(version))
    return speos_rpc_path



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a wrapped abstraction of the gRPC proto API definition and stubs."""

import logging
import os
from pathlib import Path
import subprocess
import time
from typing import TYPE_CHECKING, List, Optional, Union

import grpc
from grpc._channel import _InactiveRpcError

from ansys.api.speos.part.v1 import body_pb2, face_pb2, part_pb2
from ansys.speos.core.generic.constants import (
    DEFAULT_HOST,
    DEFAULT_PORT,
    DEFAULT_VERSION,
    MAX_CLIENT_MESSAGE_SIZE,
)
from ansys.speos.core.generic.general_methods import retrieve_speos_install_dir
from ansys.speos.core.kernel.body import BodyLink, BodyStub
from ansys.speos.core.kernel.face import FaceLink, FaceStub
from ansys.speos.core.kernel.intensity_template import (
    IntensityTemplateLink,
    IntensityTemplateStub,
)
from ansys.speos.core.kernel.job import JobLink, JobStub
from ansys.speos.core.kernel.part import PartLink, PartStub
from ansys.speos.core.kernel.scene import SceneLink, SceneStub
from ansys.speos.core.kernel.sensor_template import (
    SensorTemplateLink,
    SensorTemplateStub,
)
from ansys.speos.core.kernel.simulation_template import (
    SimulationTemplateLink,
    SimulationTemplateStub,
)
from ansys.speos.core.kernel.sop_template import (
    SOPTemplateLink,
    SOPTemplateStub,
)
from ansys.speos.core.kernel.source_template import (
    SourceTemplateLink,
    SourceTemplateStub,
)
from ansys.speos.core.kernel.spectrum import SpectrumLink, SpectrumStub
from ansys.speos.core.kernel.vop_template import (
    VOPTemplateLink,
    VOPTemplateStub,
)
from ansys.speos.core.logger import LOG as LOGGER, PySpeosCustomAdapter

if TYPE_CHECKING:  # pragma: no cover
    from ansys.platform.instancemanagement import Instance


def wait_until_healthy(channel: grpc.Channel, timeout: float):
    """
    Wait until a channel is healthy before returning.

    Parameters
    ----------
    channel : ~grpc.Channel
        Channel to wait until established and healthy.
    timeout : float
        Timeout in seconds. One attempt will be made each 100 milliseconds
        until the timeout is exceeded.

    Raises
    ------
    TimeoutError
        Raised when the total elapsed time exceeds ``timeout``.
    """
    t_max = time.time() + timeout
    while time.time() < t_max:
        try:
            grpc.channel_ready_future(channel).result(timeout=timeout)
            return True
        except (_InactiveRpcError, grpc.FutureTimeoutError):
            continue
    else:
        target_str = channel._channel.target().decode()
        raise TimeoutError(
            f"Channel health check to target '{target_str}' timed out after {timeout} seconds."
        )


class SpeosClient:
    """
    Wraps a speos gRPC connection.

    Parameters
    ----------
    host : str, optional
        Host where the server is running.
        By default, ``DEFAULT_HOST``.
    port : Union[str, int], optional
        Port number where the server is running.
        By default, ``DEFAULT_PORT``.
    channel : ~grpc.Channel, optional
        gRPC channel for server communication.
        By default, ``None``.
    message_size: int
        Maximum Message size of a newly generated channel
        By default, ``MAX_CLIENT_MESSAGE_SIZE``.
    remote_instance : ansys.platform.instancemanagement.Instance
        The corresponding remote instance when the Speos Service
        is launched through PyPIM. This instance will be deleted when calling
        :func:`SpeosClient.close <ansys.speos.core.kernel.client.SpeosClient.close >`.
    timeout : Real, optional
        Timeout in seconds to achieve the connection.
        By default, 60 seconds.
    logging_level : int, optional
        The logging level to be applied to the client.
        By default, ``INFO``.
    logging_file : Optional[str, Path]
        The file to output the log, if requested. By default, ``None``.
    speos_install_path : Optional[str, Path]
        location of Speos rpc executable
    """

    def __init__(
        self,
        host: Optional[str] = DEFAULT_HOST,
        port: Union[str, int] = DEFAULT_PORT,
        version: str = DEFAULT_VERSION,
        channel: Optional[grpc.Channel] = None,
        message_size: int = MAX_CLIENT_MESSAGE_SIZE,
        remote_instance: Optional["Instance"] = None,
        timeout: Optional[int] = 60,
        logging_level: Optional[int] = logging.INFO,
        logging_file: Optional[Union[Path, str]] = None,
        speos_install_path: Optional[Union[Path, str]] = None,
    ):
        """Initialize the ``SpeosClient`` object."""
        self._closed = False
        self._remote_instance = remote_instance
        if speos_install_path:
            speos_install_path = retrieve_speos_install_dir(speos_install_path, version)
            if os.name == "nt":
                self.__speos_exec = str(speos_install_path / "SpeosRPC_Server.exe")
            else:
                self.__speos_exec = str(speos_install_path / "SpeosRPC_Server.x")
        else:
            self.__speos_exec = None
        if not version:
            self._version = DEFAULT_VERSION
        else:
            self._version = version
        if channel:
            # Used for PyPIM when directly providing a channel
            self._channel = channel
            self._target = str(channel)
        else:
            self._host = host
            self._port = port
            self._target = f"{host}:{port}"
            self._channel = grpc.insecure_channel(
                self._target,
                options=[("grpc.max_receive_message_length", message_size)],
            )
        # do not finish initialization until channel is healthy
        wait_until_healthy(self._channel, timeout)

        # once connection with the client is established, create a logger
        self._log = LOGGER.add_instance_logger(
            name=self._target, client_instance=self, level=logging_level
        )
        if logging_file:
            if isinstance(logging_file, Path):
                logging_file = str(logging_file)
            self._log.log_to_file(filename=logging_file, level=logging_level)

        # Initialise databases
        self._faceDB = None
        self._bodyDB = None
        self._partDB = None
        self._sopTemplateDB = None
        self._vopTemplateDB = None
        self._spectrumDB = None
        self._intensityTemplateDB = None
        self._sourceTemplateDB = None
        self._sensorTemplateDB = None
        self._simulationTemplateDB = None
        self._sceneDB = None
        self._jobDB = None

    @property
    def channel(self) -> grpc.Channel:
        """The gRPC channel of this client."""
        return self._channel

    @property
    def log(self) -> PySpeosCustomAdapter:
        """The specific instance logger."""
        return self._log

    @property
    def healthy(self) -> bool:
        """Return if the client channel if healthy."""
        if self._closed:
            return False
        try:
            grpc.channel_ready_future(self.channel).result(timeout=60)
            return True
        except BaseException:
            return False

    def target(self) -> str:
        """Get the target of the channel."""
        if self._closed:
            return ""
        return self._channel._channel.target().decode()

    def faces(self) -> FaceStub:
        """Get face database access."""
        self.__closed_error()
        # connect to database
        if self._faceDB is None:
            self._faceDB = FaceStub(self._channel)
        return self._faceDB

    def bodies(self) -> BodyStub:
        """Get body database access."""
        self.__closed_error()
        # connect to database
        if self._bodyDB is None:
            self._bodyDB = BodyStub(self._channel)
        return self._bodyDB

    def parts(self) -> PartStub:
        """Get part database access."""
        self.__closed_error()
        # connect to database
        if self._partDB is None:
            self._partDB = PartStub(self._channel)
        return self._partDB

    def sop_templates(self) -> SOPTemplateStub:
        """Get sop template database access."""
        self.__closed_error()
        # connect to database
        if self._sopTemplateDB is None:
            self._sopTemplateDB = SOPTemplateStub(self._channel)
        return self._sopTemplateDB

    def vop_templates(self) -> VOPTemplateStub:
        """Get vop template database access."""
        self.__closed_error()
        # connect to database
        if self._vopTemplateDB is None:
            self._vopTemplateDB = VOPTemplateStub(self._channel)
        return self._vopTemplateDB

    def spectrums(self) -> SpectrumStub:
        """Get spectrum database access."""
        self.__closed_error()
        # connect to database
        if self._spectrumDB is None:
            self._spectrumDB = SpectrumStub(self._channel)
        return self._spectrumDB

    def intensity_templates(self) -> IntensityTemplateStub:
        """Get intensity template database access."""
        self.__closed_error()
        # connect to database
        if self._intensityTemplateDB is None:
            self._intensityTemplateDB = IntensityTemplateStub(self._channel)
        return self._intensityTemplateDB

    def source_templates(self) -> SourceTemplateStub:
        """Get source template database access."""
        self.__closed_error()
        # connect to database
        if self._sourceTemplateDB is None:
            self._sourceTemplateDB = SourceTemplateStub(self._channel)
        return self._sourceTemplateDB

    def sensor_templates(self) -> SensorTemplateStub:
        """Get sensor template database access."""
        self.__closed_error()
        # connect to database
        if self._sensorTemplateDB is None:
            self._sensorTemplateDB = SensorTemplateStub(self._channel)
        return self._sensorTemplateDB

    def simulation_templates(self) -> SimulationTemplateStub:
        """Get simulation template database access."""
        self.__closed_error()
        # connect to database
        if self._simulationTemplateDB is None:
            self._simulationTemplateDB = SimulationTemplateStub(self._channel)
        return self._simulationTemplateDB

    def scenes(self) -> SceneStub:
        """Get scene database access."""
        self.__closed_error()
        # connect to database
        if self._sceneDB is None:
            self._sceneDB = SceneStub(self._channel)
        return self._sceneDB

    def jobs(self) -> JobStub:
        """Get job database access."""
        self.__closed_error()
        # connect to database
        if self._jobDB is None:
            self._jobDB = JobStub(self._channel)
        return self._jobDB

    def __closed_error(self):
        """Check if closed."""
        if self._closed:
            raise ConnectionAbortedError()

    def __getitem__(
        self, key: str
    ) -> Union[
        SOPTemplateLink,
        VOPTemplateLink,
        SpectrumLink,
        IntensityTemplateLink,
        SourceTemplateLink,
        SensorTemplateLink,
        SimulationTemplateLink,
        SceneLink,
        JobLink,
        PartLink,
        BodyLink,
        FaceLink,
        None,
    ]:
        """Get item from key.

        Parameters
        ----------
        key : str
            Key of the item (also named guid).

        Returns
        -------
        Union[ansys.speos.core.kernel.sop_template.SOPTemplateLink, \
ansys.speos.core.kernel.vop_template.VOPTemplateLink, \
ansys.speos.core.kernel.spectrum.SpectrumLink, \
ansys.speos.core.kernel.intensity_template.IntensityTemplateLink, \
ansys.speos.core.kernel.source_template.SourceTemplateLink, \
ansys.speos.core.kernel.sensor_template.SensorTemplateLink, \
ansys.speos.core.kernel.simulation_template.SimulationTemplateLink, \
ansys.speos.core.kernel.scene.SceneLink, \
ansys.speos.core.kernel.job.JobLink, \
ansys.speos.core.kernel.part.PartLink, \
ansys.speos.core.kernel.body.BodyLink, \
ansys.speos.core.kernel.face.FaceLink, \
None]
            Link object corresponding to the key - None if no objects corresponds to the key.
        """
        self.__closed_error()
        for sop in self.sop_templates().list():
            if sop.key == key:
                return sop
        for vop in self.vop_templates().list():
            if vop.key == key:
                return vop
        for spec in self.spectrums().list():
            if spec.key == key:
                return spec
        for intens in self.intensity_templates().list():
            if intens.key == key:
                return intens
        for src in self.source_templates().list():
            if src.key == key:
                return src
        for ssr in self.sensor_templates().list():
            if ssr.key == key:
                return ssr
        for sim in self.simulation_templates().list():
            if sim.key == key:
                return sim
        for sce in self.scenes().list():
            if sce.key == key:
                return sce
        for job in self.jobs().list():
            if job.key == key:
                return job
        for part in self.parts().list():
            if part.key == key:
                return part
        for body in self.bodies().list():
            if body.key == key:
                return body
        for face in self.faces().list():
            if face.key == key:
                return face
        return None

    def get_items(
        self, keys: List[str], item_type: type
    ) -> Union[
        List[SOPTemplateLink],
        List[VOPTemplateLink],
        List[SpectrumLink],
        List[IntensityTemplateLink],
        List[SourceTemplateLink],
        List[SensorTemplateLink],
        List[SimulationTemplateLink],
        List[SceneLink],
        List[JobLink],
        List[PartLink],
        List[BodyLink],
        List[FaceLink],
    ]:
        """Get items from keys.

        Parameters
        ----------
        keys : List[str]
            Keys of the items (also named guids).
        item_type : type
            Type of items expected

        Returns
        -------
        Union[List[ansys.speos.core.kernel.sop_template.SOPTemplateLink], \
List[ansys.speos.core.kernel.vop_template.VOPTemplateLink], \
List[ansys.speos.core.kernel.spectrum.SpectrumLink], \
List[ansys.speos.core.kernel.intensity_template.IntensityTemplateLink], \
List[ansys.speos.core.kernel.source_template.SourceTemplateLink], \
List[ansys.speos.core.kernel.sensor_template.SensorTemplateLink], \
List[ansys.speos.core.kernel.simulation_template.SimulationTemplateLink], \
List[ansys.speos.core.kernel.scene.SceneLink], \
List[ansys.speos.core.kernel.job.JobLink], \
List[ansys.speos.core.kernel.part.PartLink], \
List[ansys.speos.core.kernel.body.BodyLink], \
List[ansys.speos.core.kernel.face.FaceLink]]
            List of Link objects corresponding to the keys - Empty if no objects corresponds to the
            keys.
        """
        self.__closed_error()

        if item_type == SOPTemplateLink:
            return [x for x in self.sop_templates().list() if x.key in keys]
        elif item_type == VOPTemplateLink:
            return [x for x in self.vop_templates().list() if x.key in keys]
        elif item_type == SpectrumLink:
            return [x for x in self.spectrums().list() if x.key in keys]
        elif item_type == IntensityTemplateLink:
            return [x for x in self.intensity_templates().list() if x.key in keys]
        elif item_type == SourceTemplateLink:
            return [x for x in self.source_templates().list() if x.key in keys]
        elif item_type == SensorTemplateLink:
            return [x for x in self.sensor_templates().list() if x.key in keys]
        elif item_type == SimulationTemplateLink:
            return [x for x in self.simulation_templates().list() if x.key in keys]
        elif item_type == SceneLink:
            return [x for x in self.scenes().list() if x.key in keys]
        elif item_type == JobLink:
            return [x for x in self.jobs().list() if x.key in keys]
        elif item_type == PartLink:
            guids = set(self.parts()._stubMngr.List(part_pb2.List_Request()).guids)
            return [PartLink(self.parts(), key=k) for k in keys if k in guids]
        elif item_type == BodyLink:
            guids = set(self.bodies()._stubMngr.List(body_pb2.List_Request()).guids)
            return [BodyLink(self.bodies(), key=k) for k in keys if k in guids]
        elif item_type == FaceLink:
            guids = set(self.faces()._stubMngr.List(face_pb2.List_Request()).guids)
            return [FaceLink(self.faces(), key=k) for k in keys if k in guids]
        return []

    def __repr__(self) -> str:
        """Represent the client as a string."""
        lines = []
        lines.append(f"Ansys Speos client ({hex(id(self))})")
        lines.append(f"  Target:     {self._target}")
        if self._closed:
            lines.append("  Connection: Closed")
        elif self.healthy:
            lines.append("  Connection: Healthy")
        else:
            lines.append("  Connection: Unhealthy")  # pragma: no cover
        return "\n".join(lines)

    def close(self):
        """Close the channel.

        Returns
        -------
        bool
            Information if the server instance was terminated.

        Notes
        -----
        If an instance of the Speos Service was started using
        PyPIM, this instance will be deleted.
        """
        wait_time = 0
        if self._remote_instance:
            self._remote_instance.delete()
        elif self._host in ["localhost", "0.0.0.0", "127.0.0.1"] and self.__speos_exec:
            self.__close_local_speos_rpc_server()
            while self.healthy and wait_time < 15:
                time.sleep(1)
                wait_time += 1  # takes some seconds to close rpc server
        self._channel.close()
        self._faceDB = None
        self._bodyDB = None
        self._partDB = None
        self._sopTemplateDB = None
        self._vopTemplateDB = None
        self._spectrumDB = None
        self._intensityTemplateDB = None
        self._sourceTemplateDB = None
        self._sensorTemplateDB = None
        self._simulationTemplateDB = None
        self._sceneDB = None
        self._jobDB = None
        if wait_time >= 15:
            self._closed = not self.healthy
            return self._closed
        else:
            self._closed = True
            return self._closed

    def __close_local_speos_rpc_server(self):
        command = [self.__speos_exec, "-s{}".format(self._port)]
        subprocess.run(command, check=True)



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a way to interact with Speos feature: Sensor."""

from __future__ import annotations

from difflib import SequenceMatcher
from typing import List, Mapping, Optional, Union
import uuid

import numpy as np

from ansys.api.speos.sensor.v1 import camera_sensor_pb2, common_pb2
import ansys.speos.core.generic.general_methods as general_methods
from ansys.speos.core.generic.visualization_methods import _VisualData
from ansys.speos.core.geo_ref import GeoRef
from ansys.speos.core.kernel.scene import ProtoScene
from ansys.speos.core.kernel.sensor_template import ProtoSensorTemplate
import ansys.speos.core.project as project
import ansys.speos.core.proto_message_utils as proto_message_utils


class BaseSensor:
    """Base class for Sensor.

    Parameters
    ----------
    project : ansys.speos.core.project.Project
        Project that will own the feature.
    name : str
        Name of the feature.
    description : str
        Description of the feature.
        By default, ``""``.
    metadata : Mapping[str, str]
        Metadata of the feature.
        By default, ``{}``.
    sensor_instance : ansys.api.speos.scene.v2.scene_pb2.Scene.SensorInstance, optional
        Sensor instance to provide if the feature does not has to be created from scratch
        By default, ``None``, means that the feature is created from scratch by default.

    Attributes
    ----------
    sensor_template_link : ansys.speos.core.kernel.sensor_template.SensorTemplateLink
        Link object for the sensor template in database.

    Notes
    -----
    This is a Super class, **Do not instantiate this class yourself**
    """

    def __init__(
        self,
        project: project.Project,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
        sensor_instance: Optional[ProtoScene.SensorInstance] = None,
    ) -> None:
        self._project = project
        self._name = name
        self._unique_id = None
        self._visual_data = _VisualData() if general_methods._GRAPHICS_AVAILABLE else None
        self.sensor_template_link = None
        """Link object for the sensor template in database."""
        if metadata is None:
            metadata = {}

        if sensor_instance is None:
            # Create local SensorTemplate
            self._sensor_template = ProtoSensorTemplate(
                name=name, description=description, metadata=metadata
            )

            # Create local SensorInstance
            self._sensor_instance = ProtoScene.SensorInstance(
                name=name, description=description, metadata=metadata
            )
        else:
            self._unique_id = sensor_instance.metadata["UniqueId"]
            self.sensor_template_link = self._project.client[sensor_instance.sensor_guid]
            # reset will fill _sensor_instance and _sensor_template from respectively project
            # (using _unique_id) and sensor_template_link
            self.reset()

    class WavelengthsRange:
        """Range of wavelengths.

        By default, a range from 400nm to 700nm is chosen, with a sampling of 13.

        Parameters
        ----------
        wavelengths_range : ansys.api.speos.sensor.v1.common_pb2.WavelengthsRange
            Wavelengths range protobuf object to modify.
        default_values : bool
            Uses default values when True.
        stable_ctr : bool
            Variable to indicate if usage is inside class scope

        Notes
        -----
        **Do not instantiate this class yourself**, use set_wavelengths_range method available in
        sensor classes.
        """

        def __init__(
            self,
            wavelengths_range: common_pb2.WavelengthsRange,
            default_values: bool = True,
            stable_ctr: bool = False,
        ) -> None:
            if not stable_ctr:
                msg = "WavelengthsRange class instantiated outside of class scope"
                raise RuntimeError(msg)
            self._wavelengths_range = wavelengths_range

            if default_values:
                # Default values
                self.set_start().set_end().set_sampling()

        def set_start(self, value: float = 400) -> BaseSensor.WavelengthsRange:
            """Set the minimum wavelength of the range.

            Parameters
            ----------
            value : float
                Minimum wavelength (nm).
                By default, ``400``.

            Returns
            -------
            ansys.speos.core.sensor.BaseSensor.WavelengthsRange
                WavelengthsRange.
            """
            self._wavelengths_range.w_start = value
            return self

        def set_end(self, value: float = 700) -> BaseSensor.WavelengthsRange:
            """Set the maximum wavelength of the range.

            Parameters
            ----------
            value : float
                Maximum wavelength (nm).
                By default, ``700``.

            Returns
            -------
            ansys.speos.core.sensor.BaseSensor.WavelengthsRange
                WavelengthsRange.
            """
            self._wavelengths_range.w_end = value
            return self

        def set_sampling(self, value: int = 13) -> BaseSensor.WavelengthsRange:
            """Set the sampling of wavelengths range.

            Parameters
            ----------
            value : int
                Number of wavelengths to be taken into account between the minimum and maximum
                wavelengths range.
                By default, ``13``.

            Returns
            -------
            ansys.speos.core.sensor.BaseSensor.WavelengthsRange
                WavelengthsRange.
            """
            self._wavelengths_range.w_sampling = value
            return self

    class Dimensions:
        """Dimensions of the sensor.

        By default, for both x and y axis: from -50mm to 50mm is chosen, with a sampling of 100.

        Parameters
        ----------
        sensor_dimensions : ansys.api.speos.sensor.v1.common_pb2.SensorDimensions
            SensorDimensions protobuf object to modify.
        default_values : bool
            Uses default values when True.
        stable_ctr : bool
            Variable to indicate if usage is inside class scope

        Notes
        -----
        **Do not instantiate this class yourself**, use set_dimensions method available in sensor
        classes.
        """

        def __init__(
            self,
            sensor_dimensions: common_pb2.SensorDimensions,
            default_values: bool = True,
            stable_ctr: bool = False,
        ) -> None:
            if not stable_ctr:
                msg = "Dimension class instantiated outside of class scope"
                raise RuntimeError(msg)
            self._sensor_dimensions = sensor_dimensions

            if default_values:
                # Default values
                self.set_x_start().set_x_end().set_x_sampling().set_y_start().set_y_end().set_y_sampling()

        def set_x_start(self, value: float = -50) -> BaseSensor.Dimensions:
            """Set the minimum value on x axis.

            Parameters
            ----------
            value : float
                Minimum value on x axis (mm).
                By default, ``-50``.

            Returns
            -------
            ansys.speos.core.sensor.BaseSensor.Dimensions
                Dimensions.
            """
            self._sensor_dimensions.x_start = value
            return self

        def set_x_end(self, value: float = 50) -> BaseSensor.Dimensions:
            """Set the maximum value on x axis.

            Parameters
            ----------
            value : float
                Maximum value on x axis (mm).
                By default, ``50``.

            Returns
            -------
            ansys.speos.core.sensor.BaseSensor.Dimensions
                Dimensions.
            """
            self._sensor_dimensions.x_end = value
            return self

        def set_x_sampling(self, value: int = 100) -> BaseSensor.Dimensions:
            """Set the sampling value on x axis.

            Parameters
            ----------
            value : int
                The number of pixels of the XMP map on x axis.
                By default, ``100``.

            Returns
            -------
            ansys.speos.core.sensor.BaseSensor.Dimensions
                Dimensions.
            """
            self._sensor_dimensions.x_sampling = value
            return self

        def set_y_start(self, value: float = -50) -> BaseSensor.Dimensions:
            """Set the minimum value on y axis.

            Parameters
            ----------
            value : float
                Minimum value on y axis (mm).
                By default, ``-50``.

            Returns
            -------
            ansys.speos.core.sensor.BaseSensor.Dimensions
                Dimensions.
            """
            self._sensor_dimensions.y_start = value
            return self

        def set_y_end(self, value: float = 50) -> BaseSensor.Dimensions:
            """Set the maximum value on y axis.

            Parameters
            ----------
            value : float
                Maximum value on y axis (mm).
                By default, ``50``.

            Returns
            -------
            ansys.speos.core.sensor.BaseSensor.Dimensions
                Dimensions.
            """
            self._sensor_dimensions.y_end = value
            return self

        def set_y_sampling(self, value: int = 100) -> BaseSensor.Dimensions:
            """Set the sampling value on y axis.

            Parameters
            ----------
            value : int
                The number of pixels of the XMP map on y axis.
                By default, ``100``.

            Returns
            -------
            ansys.speos.core.sensor.BaseSensor.Dimensions
                Dimensions.
            """
            self._sensor_dimensions.y_sampling = value
            return self

    class Colorimetric:
        """Type of sensor : Colorimetric.

        This kind of sensor will generate color results without any spectral data or layer
        separation in lx or W//m2.
        By default, it uses a default wavelengths range.

        Parameters
        ----------
        sensor_type_colorimetric : ansys.api.speos.sensor.v1.common_pb2.SensorTypeColorimetric
            SensorTypeColorimetric protobuf object to modify.
        default_values : bool
            Uses default values when True.
        stable_ctr : bool
            Variable to indicate if usage is inside class scope

        Notes
        -----
        **Do not instantiate this class yourself**, use set_type_colorimetric method available in
        sensor classes.
        """

        def __init__(
            self,
            sensor_type_colorimetric: common_pb2.SensorTypeColorimetric,
            default_values: bool = True,
            stable_ctr: bool = False,
        ) -> None:
            if not stable_ctr:
                msg = "Colorimetric class instantiated outside of class scope"
                raise RuntimeError(msg)
            self._sensor_type_colorimetric = sensor_type_colorimetric

            # Attribute to keep track of wavelength range object
            self._wavelengths_range = BaseSensor.WavelengthsRange(
                wavelengths_range=self._sensor_type_colorimetric.wavelengths_range,
                default_values=default_values,
                stable_ctr=stable_ctr,
            )

            if default_values:
                # Default values
                self.set_wavelengths_range()

        def set_wavelengths_range(self) -> BaseSensor.WavelengthsRange:
            """Set the range of wavelengths.

            Returns
            -------
            ansys.speos.core.sensor.BaseSensor.WavelengthsRange
                Wavelengths range.
            """
            if (
                self._wavelengths_range._wavelengths_range
                is not self._sensor_type_colorimetric.wavelengths_range
            ):
                # Happens in case of feature reset (to be sure to always modify correct data)
                self._wavelengths_range._wavelengths_range = (
                    self._sensor_type_colorimetric.wavelengths_range
                )
            return self._wavelengths_range

    class Spectral:
        """Type of sensor : Spectral.

        This kind of sensor will generate color results and spectral data separated by wavelength
        in lx or W/m2.
        By default, it uses a default wavelengths range.

        Parameters
        ----------
        sensor_type_spectral : ansys.api.speos.sensor.v1.common_pb2.SensorTypeSpectral
            SensorTypeSpectral protobuf object to modify.
        default_values : bool
            Uses default values when True.
        stable_ctr : bool
            Variable to indicate if usage is inside class scope

        Notes
        -----
        **Do not instantiate this class yourself**, use set_type_spectral method available in
        sensor classes.
        """

        def __init__(
            self,
            sensor_type_spectral: common_pb2.SensorTypeSpectral,
            default_values: bool = True,
            stable_ctr: bool = False,
        ) -> None:
            if not stable_ctr:
                msg = "Spectral class instantiated outside of class scope"
                raise RuntimeError(msg)
            self._sensor_type_spectral = sensor_type_spectral

            # Attribute to keep track of wavelength range object
            self._wavelengths_range = BaseSensor.WavelengthsRange(
                wavelengths_range=self._sensor_type_spectral.wavelengths_range,
                default_values=default_values,
                stable_ctr=stable_ctr,
            )

            if default_values:
                # Default values
                self.set_wavelengths_range()

        def set_wavelengths_range(self) -> BaseSensor.WavelengthsRange:
            """Set the range of wavelengths.

            Returns
            -------
            ansys.speos.core.sensor.BaseSensor.WavelengthsRange
                Wavelengths range.
            """
            if (
                self._wavelengths_range._wavelengths_range
                is not self._sensor_type_spectral.wavelengths_range
            ):
                self._wavelengths_range._wavelengths_range = (
                    self._sensor_type_spectral.wavelengths_range
                )
            return self._wavelengths_range

    class FaceLayer:
        """Layer composed of name and geometries.

        Parameters
        ----------
        name : str
            Name of the layer.
        geometries : List[ansys.speos.core.geo_ref.GeoRef]
            List of geometries included in this layer.

        """

        def __init__(self, name: str, geometries: List[GeoRef]) -> None:
            self.name = name
            """Name of the layer"""
            self.geometries = geometries
            """List of geometries included in this layer."""

    class LayerTypeFace:
        """Type of layer : Face.

        Includes in the result one layer per surface selected.
        By default, a filtering mode by last impact is chosen.

        Parameters
        ----------
        layer_type_face : \
        ansys.api.speos.scene.v2.scene_pb2.Scene.SensorInstance.LayerTypeFace
            LayerTypeFace protobuf object to modify.
        default_values : bool
            Uses default values when True.
        stable_ctr : bool
            Variable to indicate if usage is inside class scope

        Notes
        -----
        **Do not instantiate this class yourself**, use set_layer_type_face method available in
        sensor classes.
        """

        def __init__(
            self,
            layer_type_face: ProtoScene.SensorInstance.LayerTypeFace,
            default_values: bool = True,
            stable_ctr: bool = False,
        ) -> None:
            if not stable_ctr:
                msg = "LayerTypeFace class instantiated outside of class scope"
                raise RuntimeError(msg)

            self._layer_type_face = layer_type_face

            if default_values:
                # Default values
                self.set_sca_filtering_mode_last_impact()

        def set_sca_filtering_mode_intersected_one_time(
            self,
        ) -> BaseSensor.LayerTypeFace:
            """Set the filtering mode as intersected one time.

            Returns
            -------
            ansys.speos.core.sensor.BaseSensor.LayerTypeFace
                LayerTypeFace.
            """
            self._layer_type_face.sca_filtering_mode = (
                self._layer_type_face.EnumSCAFilteringType.IntersectedOneTime
            )
            return self

        def set_sca_filtering_mode_last_impact(
            self,
        ) -> BaseSensor.LayerTypeFace:
            """Set the filtering mode as last impact.

            Returns
            -------
            ansys.speos.core.sensor.BaseSensor.LayerTypeFace
                LayerTypeFace.
            """
            self._layer_type_face.sca_filtering_mode = (
                self._layer_type_face.EnumSCAFilteringType.LastImpact
            )
            return self

        def set_layers(self, values: List[BaseSensor.FaceLayer]) -> BaseSensor.LayerTypeFace:
            """Set the layers.

            Parameters
            ----------
            values : List[ansys.speos.core.sensor.BaseSensor.FaceLayer]
                List of layers

            Returns
            -------
            ansys.speos.core.sensor.BaseSensor.LayerTypeFace
                LayerTypeFace.
            """
            my_list = [
                ProtoScene.SensorInstance.LayerTypeFace.Layer(
                    name=layer.name,
                    geometries=ProtoScene.GeoPaths(
                        geo_paths=[gr.to_native_link() for gr in layer.geometries]
                    ),
                )
                for layer in values
            ]
            self._layer_type_face.ClearField("layers")
            self._layer_type_face.layers.extend(my_list)
            return self

    class LayerTypeSequence:
        """Type of layer : Sequence.

        Includes in the result one layer per sequence.
        By default, the sequence is defined per geometries, with a maximum number of 10 sequences.

        Parameters
        ----------
        layer_type_sequence : \
        ansys.api.speos.scene.v2.scene_pb2.Scene.SensorInstance.LayerTypeSequence
            LayerTypeSequence protobuf object to modify.
        default_values : bool
            Uses default values when True.
        stable_ctr : bool
            Variable to indicate if usage is inside class scope

        Notes
        -----
        **Do not instantiate this class yourself**, use set_layer_type_sequence method available in
        sensor classes.
        """

        def __init__(
            self,
            layer_type_sequence: ProtoScene.SensorInstance.LayerTypeSequence,
            default_values: bool = True,
            stable_ctr: bool = False,
        ) -> None:
            if not stable_ctr:
                msg = "LayerTypeSequence class instantiated outside of class scope"
                raise RuntimeError(msg)

            self._layer_type_sequence = layer_type_sequence

            if default_values:
                # Default values
                self.set_maximum_nb_of_sequence().set_define_sequence_per_geometries()

        def set_maximum_nb_of_sequence(self, value: int = 10) -> BaseSensor.LayerTypeSequence:
            """Set the maximum number of sequences.

            Parameters
            ----------
            value : int
                Maximum number of sequences.
                By default, ``10``.

            Returns
            -------
            ansys.speos.core.sensor.BaseSensor.LayerTypeSequence
                LayerTypeSequence.
            """
            self._layer_type_sequence.maximum_nb_of_sequence = value
            return self

        def set_define_sequence_per_geometries(
            self,
        ) -> BaseSensor.LayerTypeSequence:
            """Define sequence per geometries.

            Returns
            -------
            ansys.speos.core.sensor.BaseSensor.LayerTypeSequence
                LayerTypeSequence.
            """
            self._layer_type_sequence.define_sequence_per = (
                self._layer_type_sequence.EnumSequenceType.Geometries
            )
            return self

        def set_define_sequence_per_faces(self) -> BaseSensor.LayerTypeSequence:
            """Define sequence per faces.

            Returns
            -------
            ansys.speos.core.sensor.BaseSensor.LayerTypeSequence
                LayerTypeSequence.
            """
            self._layer_type_sequence.define_sequence_per = (
                self._layer_type_sequence.EnumSequenceType.Faces
            )
            return self

    class LayerTypeIncidenceAngle:
        """Type of layer : IncidenceAngle.

        Includes in the result one layer per range of incident angles.
        By default, a sampling of 9 is chosen.

        Parameters
        ----------
        layer_type_incidence_angle : \
        ansys.api.speos.scene.v2.scene_pb2.Scene.SensorInstance.LayerTypeIncidenceAngle
            LayerTypeIncidenceAngle protobuf object to modify.
        default_values : bool
            Uses default values when True.
        stable_ctr : bool
            Variable to indicate if usage is inside class scope

        Notes
        -----
        **Do not instantiate this class yourself**, use set_layer_type_incidence_angle method
        available in sensor classes.
        """

        def __init__(
            self,
            layer_type_incidence_angle: ProtoScene.SensorInstance.LayerTypeIncidenceAngle,
            default_values: bool = True,
            stable_ctr: bool = False,
        ) -> None:
            if not stable_ctr:
                msg = "LayerTypeIncidenceAngle class instantiated outside of class scope"
                raise RuntimeError(msg)

            self._layer_type_incidence_angle = layer_type_incidence_angle

            if default_values:
                # Default values
                self.set_sampling()

        def set_sampling(self, value: int = 9) -> BaseSensor.LayerTypeIncidenceAngle:
            """Set the sampling for incidence angles.

            Parameters
            ----------
            value : int
                Sampling for incidence angles.
                By default, ``9``.

            Returns
            -------
            ansys.speos.core.sensor.BaseSensor.LayerTypeIncidenceAngle
                LayerTypeIncidenceAngle.
            """
            self._layer_type_incidence_angle.sampling = value
            return self

    def _to_dict(self) -> dict:
        out_dict = {}

        # SensorInstance (= sensor guid + sensor properties)
        if self._project.scene_link and self._unique_id is not None:
            scene_data = self._project.scene_link.get()
            ssr_inst = next(
                (x for x in scene_data.sensors if x.metadata["UniqueId"] == self._unique_id),
                None,
            )
            if ssr_inst is not None:
                out_dict = proto_message_utils._replace_guids(
                    speos_client=self._project.client, message=ssr_inst
                )
            else:
                out_dict = proto_message_utils._replace_guids(
                    speos_client=self._project.client,
                    message=self._sensor_instance,
                )
        else:
            out_dict = proto_message_utils._replace_guids(
                speos_client=self._project.client, message=self._sensor_instance
            )

        if "sensor" not in out_dict.keys():
            # SensorTemplate
            if self.sensor_template_link is None:
                out_dict["sensor"] = proto_message_utils._replace_guids(
                    speos_client=self._project.client,
                    message=self._sensor_template,
                )
            else:
                out_dict["sensor"] = proto_message_utils._replace_guids(
                    speos_client=self._project.client,
                    message=self.sensor_template_link.get(),
                )

        proto_message_utils._replace_properties(json_dict=out_dict)

        return out_dict

    def get(self, key: str = "") -> str | dict:
        """Get dictionary corresponding to the project - read only.

        Parameters
        ----------
        key: str

        Returns
        -------
        str | dict
        """
        if key == "":
            return self._to_dict()
        info = proto_message_utils._value_finder_key_startswith(dict_var=self._to_dict(), key=key)
        content = list(info)
        if len(content) != 0:
            content.sort(
                key=lambda x: SequenceMatcher(None, x[0], key).ratio(),
                reverse=True,
            )
            return content[0][1]
        info = proto_message_utils._flatten_dict(dict_var=self._to_dict())
        print("Used key: {} not found in key list: {}.".format(key, info.keys()))

    def __str__(self) -> str:
        """Return the string representation of the sensor."""
        out_str = ""
        # SensorInstance (= sensor guid + sensor properties)
        if self._project.scene_link and self._unique_id is not None:
            scene_data = self._project.scene_link.get()
            ssr_inst = next(
                (x for x in scene_data.sensors if x.metadata["UniqueId"] == self._unique_id),
                None,
            )
            if ssr_inst is None:
                out_str += "local: "
        else:
            out_str += "local: "

        out_str += proto_message_utils.dict_to_str(dict=self._to_dict())

        return out_str

    def commit(self) -> BaseSensor:
        """Save feature: send the local data to the speos server database.

        Returns
        -------
        ansys.speos.core.sensor.BaseSensor
            Sensor feature.
        """
        if general_methods._GRAPHICS_AVAILABLE:
            self._visual_data.updated = False

        # The _unique_id will help to find the correct item in the scene.sensors:
        # the list of SensorInstance
        if self._unique_id is None:
            self._unique_id = str(uuid.uuid4())
            self._sensor_instance.metadata["UniqueId"] = self._unique_id

        # Save or Update the sensor template (depending on if it was already saved before)
        if self.sensor_template_link is None:
            self.sensor_template_link = self._project.client.sensor_templates().create(
                message=self._sensor_template
            )
            self._sensor_instance.sensor_guid = self.sensor_template_link.key
        elif self.sensor_template_link.get() != self._sensor_template:
            self.sensor_template_link.set(
                data=self._sensor_template
            )  # Only update if the template has changed

        # Update the scene with the sensor instance
        if self._project.scene_link:
            update_scene = True
            scene_data = self._project.scene_link.get()  # retrieve scene data

            # Look if an element corresponds to the _unique_id
            ssr_inst = next(
                (x for x in scene_data.sensors if x.metadata["UniqueId"] == self._unique_id),
                None,
            )
            if ssr_inst is not None:
                if ssr_inst != self._sensor_instance:
                    ssr_inst.CopyFrom(self._sensor_instance)  # if yes, just replace
                else:
                    update_scene = False
            else:
                scene_data.sensors.append(
                    self._sensor_instance
                )  # if no, just add it to the list of sensor instances

            if update_scene:  # Update scene only if instance has changed
                self._project.scene_link.set(data=scene_data)  # update scene data

        return self

    def reset(self) -> BaseSensor:
        """Reset feature: override local data by the one from the speos server database.

        Returns
        -------
        ansys.speos.core.sensor.BaseSensor
            Sensor feature.
        """
        # Reset sensor template
        if self.sensor_template_link is not None:
            self._sensor_template = self.sensor_template_link.get()

        # Reset sensor instance
        if self._project.scene_link is not None:
            scene_data = self._project.scene_link.get()  # retrieve scene data
            # Look if an element corresponds to the _unique_id
            ssr_inst = next(
                (x for x in scene_data.sensors if x.metadata["UniqueId"] == self._unique_id),
                None,
            )
            if ssr_inst is not None:
                self._sensor_instance = ssr_inst
        return self

    def delete(self) -> BaseSensor:
        """Delete feature: delete data from the speos server database.

        The local data are still available

        Returns
        -------
        ansys.speos.core.sensor.BaseSensor
            Sensor feature.
        """
        # Delete the sensor template
        if self.sensor_template_link is not None:
            self.sensor_template_link.delete()
            self.sensor_template_link = None

        # Reset then the sensor_guid (as the sensor template was deleted just above)
        self._sensor_instance.sensor_guid = ""

        # Remove the sensor from the scene
        scene_data = self._project.scene_link.get()  # retrieve scene data
        ssr_inst = next(
            (x for x in scene_data.sensors if x.metadata["UniqueId"] == self._unique_id),
            None,
        )
        if ssr_inst is not None:
            scene_data.sensors.remove(ssr_inst)
            self._project.scene_link.set(data=scene_data)  # update scene data

        # Reset the _unique_id
        self._unique_id = None
        self._sensor_instance.metadata.pop("UniqueId")
        return self


class SensorCamera(BaseSensor):
    """Sensor feature: Camera.

    By default, regarding inherent characteristics, a camera with mode photometric is chosen.
    By default, regarding properties, an axis system is selected to position the sensor,
    and no layer separation is chosen.

    Parameters
    ----------
    project : ansys.speos.core.project.Project
        Project that will own the feature.
    name : str
        Name of the feature.
    description : str
        Description of the feature.
        By default, ``""``.
    metadata : Optional[Mapping[str, str]]
        Metadata of the feature.
        By default, ``{}``.
    sensor_instance : ansys.api.speos.scene.v2.scene_pb2.Scene.SensorInstance, optional
        Sensor instance to provide if the feature does not has to be created from scratch
        By default, ``None``, means that the feature is created from scratch by default.
    default_values : bool
        Uses default values when True.
        By default, ``True``.
    """

    class Photometric:
        """Mode of camera sensor : Photometric.

        This allows to set every Camera Sensor parameters, including the photometric definition
        parameters.
        By default, a camera with mode color is chosen (vs monochromatic mode).

        Parameters
        ----------
        mode_photometric : ansys.api.speos.sensor.v1.camera_sensor_pb2.SensorCameraModePhotometric
            SensorCameraModePhotometric protobuf object to modify.
        default_values : bool
            Uses default values when True.
        stable_ctr : bool
            Variable to indicate if usage is inside class scope

        Notes
        -----
        **Do not instantiate this class yourself**, use set_type_photometric method available in
        sensor classes.

        """

        class Color:
            """Mode of camera sensor : Color.

            Results will be available in color according to the White Balance mode.
            By default, a balance mode none is chosen (referred as the basic conversion).

            Parameters
            ----------
            mode_color : ansys.api.speos.sensor.v1.camera_sensor_pb2.SensorCameraColorModeColor
                SensorCameraColorModeColor protobuf object to modify.
            default_values : bool
                Uses default values when True.
            stable_ctr : bool
                Variable to indicate if usage is inside class scope

            Notes
            -----
            **Do not instantiate this class yourself**, use set_mode_color method available in
            photometric class.

            """

            class BalanceModeUserWhite:
                """BalanceMode : UserWhite.

                In addition to the basic treatment, it allows to apply specific coefficients to the
                red, green, blue images.
                By default, coefficients of 1 are chosen for red, green and blue images.

                Parameters
                ----------
                balance_mode_user_white : ansys.api.speos.sensor.v1.camera_sensor_pb2.
                SensorCameraBalanceModeUserwhite
                    SensorCameraBalanceModeUserwhite protobuf object to modify.
                default_values : bool
                    Uses default values when True.
                stable_ctr : bool
                    Variable to indicate if usage is inside class scope

                Notes
                -----
                **Do not instantiate this class yourself**, use set_balance_mode_user_white method
                available in color class.

                """

                def __init__(
                    self,
                    balance_mode_user_white: camera_sensor_pb2.SensorCameraBalanceModeUserwhite,
                    default_values: bool = True,
                    stable_ctr: bool = False,
                ) -> None:
                    if not stable_ctr:
                        msg = "BalanceModeUserWhite class instantiated outside of class scope"
                        raise RuntimeError(msg)
                    self._balance_mode_user_white = balance_mode_user_white

                    if default_values:
                        # Default values
                        self.set_red_gain().set_green_gain().set_blue_gain()

                def set_red_gain(
                    self, value: float = 1
                ) -> SensorCamera.Photometric.Color.BalanceModeUserWhite:
                    """Set red gain.

                    Parameters
                    ----------
                    value : float
                        Red gain.
                        By default, ``1``.

                    Returns
                    -------
                    ansys.speos.core.sensor.SensorCamera.Photometric.Color.BalanceModeUserWhite
                        BalanceModeUserWhite.
                    """
                    self._balance_mode_user_white.red_gain = value
                    return self

                def set_green_gain(
                    self, value: float = 1
                ) -> SensorCamera.Photometric.Color.BalanceModeUserWhite:
                    """Set green gain.

                    Parameters
                    ----------
                    value : float
                        Green gain.
                        By default, ``1``.

                    Returns
                    -------
                    ansys.speos.core.sensor.SensorCamera.Photometric.Color.BalanceModeUserWhite
                        BalanceModeUserWhite.
                    """
                    self._balance_mode_user_white.green_gain = value
                    return self

                def set_blue_gain(
                    self, value: float = 1
                ) -> SensorCamera.Photometric.Color.BalanceModeUserWhite:
                    """Set blue gain.

                    Parameters
                    ----------
                    value : float
                        Blue gain.
                        By default, ``1``.

                    Returns
                    -------
                    ansys.speos.core.sensor.SensorCamera.Photometric.Color.BalanceModeUserWhite
                        BalanceModeUserWhite.
                    """
                    self._balance_mode_user_white.blue_gain = value
                    return self

            class BalanceModeDisplayPrimaries:
                """BalanceMode : DisplayPrimaries.

                Spectral results are converted in a three-channel result.
                Then a post-treatment is realized to take the distortion induced by the display
                devices into account.
                With this method, displayed results are similar to what the camera really gets.

                Parameters
                ----------
                balance_mode_display : ansys.api.speos.sensor.v1.camera_sensor_pb2.
                SensorCameraBalanceModeDisplay
                    SensorCameraBalanceModeDisplay protobuf object to modify.
                default_values : bool
                    Uses default values when True.

                Notes
                -----
                **Do not instantiate this class yourself**, use set_balance_mode_display_primaries
                method available in color class.

                """

                def __init__(
                    self,
                    balance_mode_display: camera_sensor_pb2.SensorCameraBalanceModeDisplay,
                    default_values: bool = True,
                    stable_ctr: bool = False,
                ) -> None:
                    if not stable_ctr:
                        msg = (
                            "BalanceModeDisplayPrimaries class instantiated outside of class scope"
                        )
                        raise RuntimeError(msg)

                    self._balance_mode_display = balance_mode_display

                    if default_values:
                        # Default values
                        self._balance_mode_display.SetInParent()

                def set_red_display_file_uri(
                    self, uri: str
                ) -> SensorCamera.Photometric.Color.BalanceModeDisplayPrimaries:
                    """Set the red display file.

                    Parameters
                    ----------
                    uri : str
                        Red display file.

                    Returns
                    -------
                    ansys.speos.core.sensor.SensorCamera.Photometric.Color.BalanceModeDisplayPrimaries
                        BalanceModeDisplayPrimaries.
                    """
                    self._balance_mode_display.red_display_file_uri = uri
                    return self

                def set_green_display_file_uri(
                    self, uri: str
                ) -> SensorCamera.Photometric.Color.BalanceModeDisplayPrimaries:
                    """Set the green display file.

                    Parameters
                    ----------
                    uri : str
                        Green display file.

                    Returns
                    -------
                    ansys.speos.core.sensor.SensorCamera.Photometric.Color.BalanceModeDisplayPrimaries
                        BalanceModeDisplayPrimaries.
                    """
                    self._balance_mode_display.green_display_file_uri = uri
                    return self

                def set_blue_display_file_uri(
                    self, uri: str
                ) -> SensorCamera.Photometric.Color.BalanceModeDisplayPrimaries:
                    """Set the blue display file.

                    Parameters
                    ----------
                    uri : str
                        Blue display file.

                    Returns
                    -------
                    ansys.speos.core.sensor.SensorCamera.Photometric.Color.BalanceModeDisplayPrimaries
                        BalanceModeDisplayPrimaries.
                    """
                    self._balance_mode_display.blue_display_file_uri = uri
                    return self

            def __init__(
                self,
                mode_color: camera_sensor_pb2.SensorCameraColorModeColor,
                default_values: bool = True,
                stable_ctr: bool = False,
            ) -> None:
                if not stable_ctr:
                    msg = "Color class instantiated outside of class scope"
                    raise RuntimeError(msg)
                self._mode_color = mode_color

                # Attribute gathering more complex camera balance mode
                self._mode = None

                if default_values:
                    # Default values
                    self.set_balance_mode_none()

            def set_red_spectrum_file_uri(self, uri: str) -> SensorCamera.Photometric.Color:
                """Set the red spectrum.

                Parameters
                ----------
                uri : str
                    Red spectrum file. It is expressed in a .spectrum file.

                Returns
                -------
                ansys.speos.core.sensor.SensorCamera.Photometric.Color
                    Color mode.
                """
                self._mode_color.red_spectrum_file_uri = uri
                return self

            def set_green_spectrum_file_uri(self, uri: str) -> SensorCamera.Photometric.Color:
                """Set the green spectrum.

                Parameters
                ----------
                uri : str
                    Green spectrum file. It is expressed in a .spectrum file.

                Returns
                -------
                ansys.speos.core.sensor.SensorCamera.Photometric.Color
                    Color mode.
                """
                self._mode_color.green_spectrum_file_uri = uri
                return self

            def set_blue_spectrum_file_uri(self, uri: str) -> SensorCamera.Photometric.Color:
                """Set the blue spectrum.

                Parameters
                ----------
                uri : str
                    Blue spectrum file. It is expressed in a .spectrum file.

                Returns
                -------
                ansys.speos.core.sensor.SensorCamera.Photometric.Color
                    Color mode.
                """
                self._mode_color.blue_spectrum_file_uri = uri
                return self

            def set_balance_mode_none(self) -> SensorCamera.Photometric.Color:
                """Set the balance mode as none.

                The spectral transmittance of the optical system and the spectral sensitivity for
                each channel are applied to the detected spectral image before the conversion in
                a three-channel result. This method is referred to as the basic conversion.

                Returns
                -------
                ansys.speos.core.sensor.SensorCamera.Photometric.Color
                    Color mode.
                """
                self._mode = None
                self._mode_color.balance_mode_none.SetInParent()
                return self

            def set_balance_mode_grey_world(
                self,
            ) -> SensorCamera.Photometric.Color:
                """Set the balance mode as grey world.

                The grey world assumption states that the content of the image is grey on average.
                This method converts spectral results in a three-channel result with the basic
                conversion. Then it computes and applies coefficients to the red, green and blue
                images to make sure their averages are equal.

                Returns
                -------
                ansys.speos.core.sensor.SensorCamera.Photometric.Color
                    Color mode.
                """
                self._mode = None
                self._mode_color.balance_mode_greyworld.SetInParent()
                return self

            def set_balance_mode_user_white(
                self,
            ) -> SensorCamera.Photometric.Color.BalanceModeUserWhite:
                """Set the balance mode as user white.

                In addition to the basic treatment, it allows to apply specific coefficients to the
                red, green, blue images.

                Returns
                -------
                ansys.speos.core.sensor.SensorCamera.Photometric.Color.BalanceModeUserWhite
                    Balance UserWhite mode.
                """
                if self._mode is None and self._mode_color.HasField("balance_mode_userwhite"):
                    # Happens in case of project created via load of speos file
                    self._mode = SensorCamera.Photometric.Color.BalanceModeUserWhite(
                        balance_mode_user_white=self._mode_color.balance_mode_userwhite,
                        default_values=False,
                        stable_ctr=True,
                    )
                elif not isinstance(
                    self._mode, SensorCamera.Photometric.Color.BalanceModeUserWhite
                ):
                    # if the _mode is not BalanceModeUserWhite then we create a new type.
                    self._mode = SensorCamera.Photometric.Color.BalanceModeUserWhite(
                        balance_mode_user_white=self._mode_color.balance_mode_userwhite,
                        stable_ctr=True,
                    )
                elif (
                    self._mode._balance_mode_user_white
                    is not self._mode_color.balance_mode_userwhite
                ):
                    # Happens in case of feature reset (to be sure to always modify correct data)
                    self._mode._balance_mode_user_white = self._mode_color.balance_mode_userwhite
                return self._mode

            def set_balance_mode_display_primaries(
                self,
            ) -> SensorCamera.Photometric.Color.BalanceModeDisplayPrimaries:
                """Set the balance mode as display primaries.

                Spectral results are converted in a three-channel result.
                Then a post-treatment is realized to take the distortion induced by the display
                devices into account. With this method, displayed results are similar to what the
                camera really gets.

                Returns
                -------
                ansys.speos.core.sensor.SensorCamera.Photometric.Color.BalanceModeDisplayPrimaries
                    Balance DisplayPrimaries mode.
                """
                if self._mode is None and self._mode_color.HasField("balance_mode_display"):
                    # Happens in case of project created via load of speos file
                    self._mode = SensorCamera.Photometric.Color.BalanceModeDisplayPrimaries(
                        balance_mode_display=self._mode_color.balance_mode_display,
                        default_values=False,
                        stable_ctr=True,
                    )
                elif not isinstance(
                    self._mode, SensorCamera.Photometric.Color.BalanceModeDisplayPrimaries
                ):
                    # if the _mode is not BalanceModeDisplayPrimaries then we create a new type.
                    self._mode = SensorCamera.Photometric.Color.BalanceModeDisplayPrimaries(
                        balance_mode_display=self._mode_color.balance_mode_display,
                        stable_ctr=True,
                    )
                elif self._mode._balance_mode_display is not self._mode_color.balance_mode_display:
                    # Happens in case of feature reset (to be sure to always modify correct data)
                    self._mode._balance_mode_display = self._mode_color.balance_mode_display
                return self._mode

        def __init__(
            self,
            mode_photometric: camera_sensor_pb2.SensorCameraModePhotometric,
            camera_props: ProtoScene.SensorInstance.CameraProperties,
            default_values: bool = True,
            stable_ctr: bool = False,
        ) -> None:
            if not stable_ctr:
                msg = "Photometric class instantiated outside of class scope"
                raise RuntimeError(msg)
            self._mode_photometric = mode_photometric
            self._camera_props = camera_props

            # Attribute gathering more complex camera color mode
            self._mode = None

            # Attribute to keep track of wavelength range object
            self._wavelengths_range = SensorCamera.WavelengthsRange(
                wavelengths_range=self._mode_photometric.wavelengths_range,
                stable_ctr=stable_ctr,
            )

            if default_values:
                # Default values
                self.set_acquisition_integration().set_acquisition_lag_time().set_gamma_correction().set_png_bits_16().set_mode_color()
                self.set_wavelengths_range()
                # Default values properties
                self.set_layer_type_none()

        def set_acquisition_integration(self, value: float = 0.01) -> SensorCamera.Photometric:
            """Set the acquisition integration value.

            Parameters
            ----------
            value : float
                Acquisition integration value (s).
                By default, ``0.01``.

            Returns
            -------
            ansys.speos.core.sensor.SensorCamera.Photometric
                Photometric mode.
            """
            self._mode_photometric.acquisition_integration = value
            return self

        def set_acquisition_lag_time(self, value: float = 0.0) -> SensorCamera.Photometric:
            """Set the acquisition lag time value.

            Parameters
            ----------
            value : float
                Acquisition lag time value (s).
                By default, ``0.0``.

            Returns
            -------
            ansys.speos.core.sensor.SensorCamera.Photometric
                Photometric mode.
            """
            self._mode_photometric.acquisition_lag_time = value
            return self

        def set_transmittance_file_uri(self, uri: str) -> SensorCamera.Photometric:
            """Set the transmittance file.

            Parameters
            ----------
            uri : str
                Amount of light of the source that passes through the lens and reaches the sensor.
                The transmittance is expressed in a .spectrum file.

            Returns
            -------
            ansys.speos.core.sensor.SensorCamera.Photometric
                Photometric mode.
            """
            self._mode_photometric.transmittance_file_uri = uri
            return self

        def set_gamma_correction(self, value: float = 2.2) -> SensorCamera.Photometric:
            """Set the gamma correction.

            Parameters
            ----------
            value : float
                Compensation of the curve before the display on the screen.
                By default, ``2.2``.

            Returns
            -------
            ansys.speos.core.sensor.SensorCamera.Photometric
                Photometric mode.
            """
            self._mode_photometric.gamma_correction = value
            return self

        def set_png_bits_08(self) -> SensorCamera.Photometric:
            """Choose 08-bits for png.

            Returns
            -------
            ansys.speos.core.sensor.SensorCamera.Photometric
                Photometric mode.
            """
            self._mode_photometric.png_bits = camera_sensor_pb2.EnumSensorCameraPNGBits.PNG_08
            return self

        def set_png_bits_10(self) -> SensorCamera.Photometric:
            """Choose 10-bits for png.

            Returns
            -------
            ansys.speos.core.sensor.SensorCamera.Photometric
                Photometric mode.
            """
            self._mode_photometric.png_bits = camera_sensor_pb2.EnumSensorCameraPNGBits.PNG_10
            return self

        def set_png_bits_12(self) -> SensorCamera.Photometric:
            """Choose 12-bits for png.

            Returns
            -------
            ansys.speos.core.sensor.SensorCamera.Photometric
                Photometric mode.
            """
            self._mode_photometric.png_bits = camera_sensor_pb2.EnumSensorCameraPNGBits.PNG_12
            return self

        def set_png_bits_16(self) -> SensorCamera.Photometric:
            """Choose 16-bits for png.

            Returns
            -------
            ansys.speos.core.sensor.SensorCamera.Photometric
                Photometric mode.
            """
            self._mode_photometric.png_bits = camera_sensor_pb2.EnumSensorCameraPNGBits.PNG_16
            return self

        def set_wavelengths_range(self) -> BaseSensor.WavelengthsRange:
            """Set the range of wavelengths.

            Returns
            -------
            ansys.speos.core.sensor.BaseSensor.WavelengthsRange
                Wavelengths range.
            """
            if (
                self._wavelengths_range._wavelengths_range
                is not self._mode_photometric.wavelengths_range
            ):
                # Happens in case of feature reset (to be sure to always modify correct data)
                self._wavelengths_range._wavelengths_range = (
                    self._mode_photometric.wavelengths_range
                )
            return self._wavelengths_range

        def set_mode_monochromatic(self, spectrum_file_uri: str) -> SensorCamera.Photometric:
            """Set the monochromatic mode.

            Results will be available in grey scale.

            Parameters
            ----------
            spectrum_file_uri : str
                Spectrum file uri.

            Returns
            -------
            ansys.speos.core.sensor.SensorCamera.Photometric
                Photometric mode.
            """
            self._mode = None
            self._mode_photometric.color_mode_monochromatic.spectrum_file_uri = spectrum_file_uri
            return self

        def set_mode_color(self) -> SensorCamera.Photometric.Color:
            """Set the color mode.

            Results will be available in color.

            Returns
            -------
            ansys.speos.core.sensor.SensorCamera.Photometric.Color
                Color mode.
            """
            if self._mode is None and self._mode_photometric.HasField("color_mode_color"):
                # Happens in case of project created via load of speos file
                self._mode = SensorCamera.Photometric.Color(
                    mode_color=self._mode_photometric.color_mode_color,
                    default_values=False,
                    stable_ctr=True,
                )
            elif not isinstance(self._mode, SensorCamera.Photometric.Color):
                # if the _mode is not Color then we create a new type.
                self._mode = SensorCamera.Photometric.Color(
                    mode_color=self._mode_photometric.color_mode_color,
                    stable_ctr=True,
                )
            elif self._mode._mode_color is not self._mode_photometric.color_mode_color:
                # Happens in case of feature reset (to be sure to always modify correct data)
                self._mode._mode_color = self._mode_photometric.color_mode_color
            return self._mode

        def set_trajectory_file_uri(self, uri: str) -> SensorCamera.Photometric:
            """Set the trajectory file.

            Parameters
            ----------
            uri : str
                Trajectory file, used to define the position and orientations of the Camera sensor
                in time.

            Returns
            -------
            ansys.speos.core.sensor.SensorCamera.Photometric
                Photometric mode.
            """
            self._camera_props.trajectory_file_uri = uri
            return self

        def set_layer_type_none(self) -> SensorCamera.Photometric:
            """Set no layer separation: includes the simulation's results in one layer.

            Returns
            -------
            ansys.speos.core.sensor.SensorCamera.Photometric
                Photometric mode.
            """
            self._camera_props.layer_type_none.SetInParent()
            return self

        def set_layer_type_source(self) -> SensorCamera.Photometric:
            """Set layer separation by source: includes one layer per active source in the result.

            Returns
            -------
            ansys.speos.core.sensor.SensorCamera.Photometric
                Photometric mode.
            """
            self._camera_props.layer_type_source.SetInParent()
            return self

    def __init__(
        self,
        project: project.Project,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
        sensor_instance: Optional[ProtoScene.SensorInstance] = None,
        default_values: bool = True,
    ) -> None:
        if metadata is None:
            metadata = {}

        super().__init__(
            project=project,
            name=name,
            description=description,
            metadata=metadata,
            sensor_instance=sensor_instance,
        )

        # Attribute gathering more complex camera mode
        self._type = None
        if default_values:
            # Default values template
            self.set_focal_length().set_imager_distance().set_f_number().set_horz_pixel()
            self.set_vert_pixel().set_width().set_height().set_mode_photometric()
            # Default values properties
            self.set_axis_system()

    @property
    def visual_data(self) -> _VisualData:
        """Property containing camera sensor visualization data.

        Returns
        -------
        BaseSensor.VisualData
            Instance of VisualData Class for pyvista.PolyData of feature faces, coordinate_systems.

        """
        if self._visual_data.updated:
            return self._visual_data
        else:
            feature_pos_info = self.get(key="axis_system")
            feature_camera_pos = np.array(feature_pos_info[:3])
            feature_camera_x_dir = np.array(feature_pos_info[3:6])
            feature_camera_y_dir = np.array(feature_pos_info[6:9])
            feature_camera_z_dir = np.array(feature_pos_info[9:12])
            feature_width = float(self.get(key="width"))
            feature_height = float(self.get(key="height"))
            feature_camera_focal = float(self.get(key="focal_length"))
            feature_camera_image_dis = float(self.get(key="imager_distance"))

            # camera radiance sensor
            p1 = (
                feature_camera_pos
                + feature_camera_x_dir * feature_width / 2.0
                + feature_camera_y_dir * feature_height / 2.0
                + feature_camera_z_dir * feature_camera_image_dis
            )
            p2 = (
                feature_camera_pos
                + feature_camera_x_dir * feature_width / 2.0
                - feature_camera_y_dir * feature_height / 2.0
                + feature_camera_z_dir * feature_camera_image_dis
            )
            p3 = (
                feature_camera_pos
                - feature_camera_x_dir * feature_width / 2.0
                + feature_camera_y_dir * feature_height / 2.0
                + feature_camera_z_dir * feature_camera_image_dis
            )
            p4 = (
                feature_camera_pos
                - feature_camera_x_dir * feature_width / 2.0
                - feature_camera_y_dir * feature_height / 2.0
                + feature_camera_z_dir * feature_camera_image_dis
            )
            self._visual_data.add_data_rectangle([p1, p2, p3])

            p5 = feature_camera_pos + feature_camera_z_dir * (
                feature_camera_image_dis - feature_camera_focal
            )
            self._visual_data.add_data_triangle([p1, p2, p5])
            self._visual_data.add_data_triangle([p3, p4, p5])
            self._visual_data.add_data_triangle([p1, p3, p5])
            self._visual_data.add_data_triangle([p2, p4, p5])

            # NOTE: camera object field to be added
            # current gRPC service not available to get object field open angle
            # camera_object_field_radius = 500
            # camera_object_field_data = pv.Sphere(
            #     radius=camera_object_field_radius,
            #     center=feature_camera_pos,
            #     direction=feature_camera_y_dir,
            #     theta_resolution=30,
            #     phi_resolution=30,
            #     start_theta=0.0,
            #     end_theta=60.0,
            #     start_phi=90,
            #     end_phi=135,
            # )
            # camera_object_field_data = camera_object_field_data.rotate_vector(
            #     vector=feature_camera_y_dir, angle=-30, point=feature_camera_pos
            # )
            # self._visual_data.data = self._visual_data.data.append_polydata(
            #     camera_object_field_data
            # )

            # camera axis system
            self._visual_data.coordinates.origin = feature_camera_pos
            self._visual_data.coordinates.x_axis = (
                feature_camera_x_dir * max(feature_width, feature_height) / 4.0
            )
            self._visual_data.coordinates.y_axis = (
                feature_camera_y_dir * max(feature_width, feature_height) / 4.0
            )
            self._visual_data.coordinates.z_axis = (
                feature_camera_z_dir * max(feature_width, feature_height) / 4.0
            )

            self._visual_data.updated = True
            return self._visual_data

    @property
    def photometric(self) -> Union[SensorCamera.Photometric, None]:
        """Property containing the instance of SensorCamera.Photometric used to build the sensor.

        Returns
        -------
        Union[ansys.speos.core.sensor.SensorCamera.Photometric, None]
            Photometric class instance if it exists

        """
        return self._type

    def set_focal_length(self, value: float = 5.0) -> SensorCamera:
        """Set the focal length.

        Parameters
        ----------
        value : float
            Distance between the center of the optical system and the focus. (mm)
            By default, ``5.0``.

        Returns
        -------
        ansys.speos.core.sensor.SensorCamera
            Camera feature
        """
        self._sensor_template.camera_sensor_template.focal_length = value
        return self

    def set_imager_distance(self, value: float = 10) -> SensorCamera:
        """Set the imager distance.

        Parameters
        ----------
        value : float
            Imager distance (mm). The imager is located at the focal point.
            The Imager distance has no impact on the result.
            By default, ``10``.

        Returns
        -------
        ansys.speos.core.sensor.SensorCamera
            Camera feature
        """
        self._sensor_template.camera_sensor_template.imager_distance = value
        return self

    def set_f_number(self, value: float = 20) -> SensorCamera:
        """Set the f number.

        Parameters
        ----------
        value : float
            F-number represents the aperture of the front lens.
            F number has no impact on the result.
            By default, ``20``.

        Returns
        -------
        ansys.speos.core.sensor.SensorCamera
            Camera feature
        """
        self._sensor_template.camera_sensor_template.f_number = value
        return self

    def set_distortion_file_uri(self, uri: str) -> SensorCamera:
        """Set the distortion file.

        Parameters
        ----------
        uri : str
            Optical aberration that deforms and bends straight lines. The distortion is expressed in
            a .OPTDistortion file.

        Returns
        -------
        ansys.speos.core.sensor.SensorCamera
            Camera feature
        """
        self._sensor_template.camera_sensor_template.distortion_file_uri = uri
        return self

    def set_horz_pixel(self, value: int = 640) -> SensorCamera:
        """Set the horizontal pixels number corresponding to the camera resolution.

        Parameters
        ----------
        value : int
            The horizontal pixels number corresponding to the camera resolution.
            By default, ``640``.

        Returns
        -------
        ansys.speos.core.sensor.SensorCamera
            Camera feature
        """
        self._sensor_template.camera_sensor_template.horz_pixel = value
        return self

    def set_vert_pixel(self, value: int = 480) -> SensorCamera:
        """Set the vertical pixels number corresponding to the camera resolution.

        Parameters
        ----------
        value : int
            The vertical pixels number corresponding to the camera resolution.
            By default, ``480``.

        Returns
        -------
        ansys.speos.core.sensor.SensorCamera
            Camera feature
        """
        self._sensor_template.camera_sensor_template.vert_pixel = value
        return self

    def set_width(self, value: float = 5.0) -> SensorCamera:
        """Set the width of the sensor.

        Parameters
        ----------
        value : float
            Sensor's width (mm).
            By default, ``5.0``.

        Returns
        -------
        ansys.speos.core.sensor.SensorCamera
            Camera feature
        """
        self._sensor_template.camera_sensor_template.width = value
        return self

    def set_height(self, value: float = 5.0) -> SensorCamera:
        """Set the height of the sensor.

        Parameters
        ----------
        value : float
            Sensor's height (mm).
            By default, ``5.0``.

        Returns
        -------
        ansys.speos.core.sensor.SensorCamera
            Camera feature
        """
        self._sensor_template.camera_sensor_template.height = value
        return self

    def set_mode_geometric(self) -> SensorCamera:
        """Set mode geometric for the camera sensor.

        This is a simplified version of the Camera Sensor.

        Returns
        -------
        ansys.speos.core.sensor.SensorCamera
            Geometric Camera feature
        """
        self._type = None
        self._sensor_template.camera_sensor_template.sensor_mode_geometric.SetInParent()
        return self

    def set_mode_photometric(self) -> SensorCamera.Photometric:
        """Set mode photometric for the camera sensor.

        This allows setting every Camera Sensor parameter, including the photometric definition
        parameters.

        Returns
        -------
        ansys.speos.core.sensor.SensorCamera.Photometric
            Photometric Camera Sensor feature
        """
        if self._type is None and self._sensor_template.camera_sensor_template.HasField(
            "sensor_mode_photometric"
        ):
            # Happens in case of project created via load of speos file
            self._type = SensorCamera.Photometric(
                mode_photometric=self._sensor_template.camera_sensor_template.sensor_mode_photometric,
                camera_props=self._sensor_instance.camera_properties,
                default_values=False,
                stable_ctr=True,
            )
        elif not isinstance(self._type, SensorCamera.Photometric):
            # if the _type is not Photometric then we create a new type.
            self._type = SensorCamera.Photometric(
                mode_photometric=self._sensor_template.camera_sensor_template.sensor_mode_photometric,
                camera_props=self._sensor_instance.camera_properties,
                stable_ctr=True,
            )
        elif (
            self._type._mode_photometric
            is not self._sensor_template.camera_sensor_template.sensor_mode_photometric
        ):
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._type._mode_photometric = (
                self._sensor_template.camera_sensor_template.sensor_mode_photometric
            )
        return self._type

    def set_axis_system(self, axis_system: Optional[List[float]] = None) -> SensorCamera:
        """Set the position of the sensor.

        Parameters
        ----------
        axis_system : Optional[List[float]]
            Position of the sensor [Ox Oy Oz Xx Xy Xz Yx Yy Yz Zx Zy Zz].
            By default, ``[0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1]``.

        Returns
        -------
        ansys.speos.core.sensor.SensorCamera
            Camera Sensor feature
        """
        if axis_system is None:
            axis_system = [0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1]
        self._sensor_instance.camera_properties.axis_system[:] = axis_system
        return self


class SensorIrradiance(BaseSensor):
    """Sensor feature: Irradiance.

    By default, regarding inherent characteristics, an irradiance sensor of type photometric and
    illuminance type planar is chosen. By default, regarding properties, an axis system is
    selected to position the sensor, no layer separation and no ray file generation are chosen.

    Parameters
    ----------
    project : ansys.speos.core.project.Project
        Project that will own the feature.
    name : str
        Name of the feature.
    description : str
        Description of the feature.
        By default, ``""``.
    metadata : Optional[Mapping[str, str]]
        Metadata of the feature.
        By default, ``{}``.
    sensor_instance : ansys.api.speos.scene.v2.scene_pb2.Scene.SensorInstance, optional
        Sensor instance to provide if the feature does not has to be created from scratch
        By default, ``None``, means that the feature is created from scratch by default.
    default_values : bool
        Uses default values when True.
        By default, ``True``.
    """

    def __init__(
        self,
        project: project.Project,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
        sensor_instance: Optional[ProtoScene.SensorInstance] = None,
        default_values: bool = True,
    ) -> None:
        if metadata is None:
            metadata = {}

        super().__init__(
            project=project,
            name=name,
            description=description,
            metadata=metadata,
            sensor_instance=sensor_instance,
        )

        # Attribute gathering more complex irradiance type
        self._type = None

        # Attribute gathering more complex layer type
        self._layer_type = None

        # Attribute to keep track of sensor dimensions object
        self._sensor_dimensions = self.Dimensions(
            sensor_dimensions=self._sensor_template.irradiance_sensor_template.dimensions,
            default_values=default_values,
            stable_ctr=True,
        )

        if default_values:
            # Default values template
            self.set_type_photometric().set_illuminance_type_planar()
            # Default values properties
            self.set_axis_system().set_ray_file_type_none().set_layer_type_none()

    @property
    def visual_data(self) -> _VisualData:
        """Property containing irradiance sensor visualization data.

        Returns
        -------
        BaseSensor.VisualData
            Instance of VisualData Class for pyvista.PolyData of feature faces, coordinate_systems.

        """
        if self._visual_data.updated is True:
            return self._visual_data
        else:
            feature_pos_info = self.get(key="axis_system")
            feature_irradiance_pos = np.array(feature_pos_info[:3])
            feature_irradiance_x_dir = np.array(feature_pos_info[3:6])
            feature_irradiance_y_dir = np.array(feature_pos_info[6:9])
            if self._sensor_instance.irradiance_properties.integration_direction:
                feature_irradiance_z_dir = np.array(
                    self._sensor_instance.irradiance_properties.integration_direction
                )
            else:
                feature_irradiance_z_dir = np.array(feature_pos_info[9:12])
            feature_x_start = float(self.get(key="x_start"))
            feature_x_end = float(self.get(key="x_end"))
            feature_y_start = float(self.get(key="y_start"))
            feature_y_end = float(self.get(key="y_end"))

            # irradiance sensor
            p1 = (
                feature_irradiance_pos
                + feature_irradiance_x_dir * feature_x_end
                + feature_irradiance_y_dir * feature_y_end
            )
            p2 = (
                feature_irradiance_pos
                + feature_irradiance_x_dir * feature_x_start
                + feature_irradiance_y_dir * feature_y_start
            )
            p3 = (
                feature_irradiance_pos
                + feature_irradiance_x_dir * feature_x_end
                + feature_irradiance_y_dir * feature_y_start
            )
            self._visual_data.add_data_rectangle([p1, p2, p3])

            # irradiance direction
            self._visual_data.coordinates.origin = feature_irradiance_pos
            self._visual_data.coordinates.x_axis = feature_irradiance_x_dir * max(
                feature_y_end - feature_y_start, feature_x_end - feature_x_start
            )
            self._visual_data.coordinates.y_axis = feature_irradiance_y_dir * max(
                feature_y_end - feature_y_start, feature_x_end - feature_x_start
            )
            self._visual_data.coordinates.z_axis = feature_irradiance_z_dir * max(
                feature_y_end - feature_y_start, feature_x_end - feature_x_start
            )

            self._visual_data.updated = True
            return self._visual_data

    @property
    def dimensions(self) -> BaseSensor.Dimensions:
        """Property containing all options in regard to the Dimensions sensor properties.

        Returns
        -------
        ansys.speos.core.sensor.BaseSensor.Dimensions
            Instance of Dimensions Class for this sensor feature
        """
        return self._sensor_dimensions

    @property
    def type(self) -> str:
        """Type of sensor.

        Returns
        -------
        str
            Sensor type as string
        """
        if type(self._type) is str:
            return self._type
        elif isinstance(self._type, BaseSensor.Colorimetric):
            return "Colorimetric"
        elif isinstance(self._type, BaseSensor.Spectral):
            return "Spectral"
        else:
            return self._type

    @property
    def colorimetric(self) -> Union[None, BaseSensor.Colorimetric]:
        """Property containing all options in regard to the Colorimetric sensor properties.

        Returns
        -------
        Union[None, ansys.speos.core.sensor.BaseSensor.Colorimetric]
            Instance of Colorimetric Class for this sensor feature
        """
        if isinstance(self._type, BaseSensor.Colorimetric):
            return self._type
        else:
            return None

    @property
    def spectral(self) -> Union[None, BaseSensor.Spectral]:
        """Property containing all options in regard to the Spectral sensor properties.

        Returns
        -------
        Union[None, ansys.speos.core.sensor.BaseSensor.Spectral]
            Instance of Spectral Class for this sensor feature
        """
        if isinstance(self._type, BaseSensor.Spectral):
            return self._type
        else:
            return None

    @property
    def layer(
        self,
    ) -> Union[
        None,
        SensorIrradiance,
        BaseSensor.LayerTypeFace,
        BaseSensor.LayerTypeSequence,
        BaseSensor.LayerTypeIncidenceAngle,
    ]:
        """Property containing all options in regard to the layer separation properties.

        Returns
        -------
        Union[\
            None,\
            ansys.speos.core.sensor.SensorIrradiance,\
            ansys.speos.core.sensor.BaseSensor.LayerTypeFace,\
            ansys.speos.core.sensor.BaseSensor.LayerTypeSequence,\
            ansys.speos.core.sensor.BaseSensor.LayerTypeIncidenceAngle\
        ]
            Instance of Layertype Class for this sensor feature
        """
        return self._layer_type

    def set_dimensions(self) -> BaseSensor.Dimensions:
        """Set the dimensions of the sensor.

        Returns
        -------
        ansys.speos.core.sensor.BaseSensor.Dimensions
            Dimension class
        """
        if (
            self._sensor_dimensions._sensor_dimensions
            is not self._sensor_template.irradiance_sensor_template.dimensions
        ):
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._sensor_dimensions._sensor_dimensions = (
                self._sensor_template.irradiance_sensor_template.dimensions
            )
        return self._sensor_dimensions

    def set_type_photometric(self) -> SensorIrradiance:
        """Set type photometric.

        The sensor considers the visible spectrum and gets the results in lm/m2 or lx.

        Returns
        -------
        ansys.speos.core.sensor.SensorIrradiance
            Irradiance sensor
        """
        self._sensor_template.irradiance_sensor_template.sensor_type_photometric.SetInParent()
        self._type = "Photometric"
        return self

    def set_type_colorimetric(self) -> BaseSensor.Colorimetric:
        """Set type colorimetric.

        The sensor will generate color results without any spectral data or layer separation
        in lx or W//m2.

        Returns
        -------
        ansys.speos.core.sensor.BaseSensor.Colorimetric
            Colorimetric type.
        """
        if self._type is None and self._sensor_template.irradiance_sensor_template.HasField(
            "sensor_type_colorimetric"
        ):
            # Happens in case of project created via load of speos file
            self._type = BaseSensor.Colorimetric(
                sensor_type_colorimetric=self._sensor_template.irradiance_sensor_template.sensor_type_colorimetric,
                default_values=False,
                stable_ctr=True,
            )
        elif not isinstance(self._type, BaseSensor.Colorimetric):
            # if the _type is not Colorimetric then we create a new type.
            self._type = BaseSensor.Colorimetric(
                sensor_type_colorimetric=self._sensor_template.irradiance_sensor_template.sensor_type_colorimetric,
                stable_ctr=True,
            )
        elif (
            self._type._sensor_type_colorimetric
            is not self._sensor_template.irradiance_sensor_template.sensor_type_colorimetric
        ):
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._type._sensor_type_colorimetric = (
                self._sensor_template.irradiance_sensor_template.sensor_type_colorimetric
            )
        return self._type

    def set_type_radiometric(self) -> SensorIrradiance:
        """Set type radiometric.

        The sensor considers the entire spectrum and gets the results in W/m2.

        Returns
        -------
        ansys.speos.core.sensor.SensorIrradiance
            Irradiance sensor.
        """
        self._sensor_template.irradiance_sensor_template.sensor_type_radiometric.SetInParent()
        self._type = "Radiometric"
        return self

    def set_type_spectral(self) -> BaseSensor.Spectral:
        """Set type spectral.

        The sensor will generate color results and spectral data separated by wavelength
        in lx or W/m2.

        Returns
        -------
        ansys.speos.core.sensor.BaseSensor.Spectral
            Spectral type.
        """
        if self._type is None and self._sensor_template.irradiance_sensor_template.HasField(
            "sensor_type_spectral"
        ):
            # Happens in case of project created via load of speos file
            self._type = BaseSensor.Spectral(
                sensor_type_spectral=self._sensor_template.irradiance_sensor_template.sensor_type_spectral,
                default_values=False,
                stable_ctr=True,
            )
        elif not isinstance(self._type, BaseSensor.Spectral):
            # if the _type is not Spectral then we create a new type.
            self._type = BaseSensor.Spectral(
                sensor_type_spectral=self._sensor_template.irradiance_sensor_template.sensor_type_spectral,
                stable_ctr=True,
            )
        elif (
            self._type._sensor_type_spectral
            is not self._sensor_template.irradiance_sensor_template.sensor_type_spectral
        ):
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._type._sensor_type_spectral = (
                self._sensor_template.irradiance_sensor_template.sensor_type_spectral
            )
        return self._type

    def set_illuminance_type_planar(
        self, integration_direction: Optional[List[float]] = None
    ) -> SensorIrradiance:
        """Set illuminance type planar.

        Parameters
        ----------
        integration_direction : List[float], optional
            Sensor global integration direction [x,y,z].
            The integration direction must be set in the anti-rays direction to integrate their
            signal.
            By default, ``None``. None means that the Z axis of axis_system is taken.

        Returns
        -------
        ansys.speos.core.sensor.SensorIrradiance
            Irradiance sensor.

        Notes
        -----
        Contrary to any visualization of integration directions within Speos Software or its
        documentation, the integration direction must be set in the anti-rays direction to integrate
        their signal.
        """
        self._sensor_template.irradiance_sensor_template.illuminance_type_planar.SetInParent()
        if not integration_direction:
            self._sensor_instance.irradiance_properties.ClearField("integration_direction")
        else:
            self._sensor_instance.irradiance_properties.integration_direction[:] = (
                integration_direction
            )
        return self

    def set_illuminance_type_radial(self) -> SensorIrradiance:
        """Set illuminance type radial.

        Returns
        -------
        ansys.speos.core.sensor.SensorIrradiance
            Irradiance sensor.
        """
        self._sensor_template.irradiance_sensor_template.illuminance_type_radial.SetInParent()
        return self

    def set_illuminance_type_hemispherical(self) -> SensorIrradiance:
        """Set illuminance type hemispherical.

        Returns
        -------
        ansys.speos.core.sensor.SensorIrradiance
            Irradiance sensor.
        """
        self._sensor_template.irradiance_sensor_template.illuminance_type_hemispherical.SetInParent()
        return self

    def set_illuminance_type_cylindrical(self) -> SensorIrradiance:
        """Set illuminance type cylindrical.

        Returns
        -------
        ansys.speos.core.sensor.SensorIrradiance
            Irradiance sensor.
        """
        self._sensor_template.irradiance_sensor_template.illuminance_type_cylindrical.SetInParent()
        return self

    def set_illuminance_type_semi_cylindrical(
        self, integration_direction: Optional[List[float]] = None
    ) -> SensorIrradiance:
        """Set illuminance type semi cylindrical.

        Parameters
        ----------
        integration_direction : List[float], optional
            Sensor global integration direction [x,y,z].
            By default, ``None``. None means that the Z axis of axis_system is taken.

        Returns
        -------
        ansys.speos.core.sensor.SensorIrradiance
            Irradiance sensor.
        """
        self._sensor_template.irradiance_sensor_template.illuminance_type_semi_cylindrical.SetInParent()
        if not integration_direction:
            self._sensor_instance.irradiance_properties.ClearField("integration_direction")
        else:
            self._sensor_instance.irradiance_properties.integration_direction[:] = (
                integration_direction
            )
        return self

    def set_axis_system(self, axis_system: Optional[List[float]] = None) -> SensorIrradiance:
        """Set position of the sensor.

        Parameters
        ----------
        axis_system : Optional[List[float]]
            Position of the sensor [Ox Oy Oz Xx Xy Xz Yx Yy Yz Zx Zy Zz].
            By default, ``[0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1]``.

        Returns
        -------
        ansys.speos.core.sensor.SensorIrradiance
            Irradiance sensor.
        """
        if axis_system is None:
            axis_system = [0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1]
        self._sensor_instance.irradiance_properties.axis_system[:] = axis_system
        return self

    def set_ray_file_type_none(self) -> SensorIrradiance:
        """Set no ray file generation.

        Returns
        -------
        ansys.speos.core.sensor.SensorIrradiance
            Irradiance sensor.
        """
        self._sensor_instance.irradiance_properties.ray_file_type = (
            ProtoScene.SensorInstance.EnumRayFileType.RayFileNone
        )
        return self

    def set_ray_file_type_classic(self) -> SensorIrradiance:
        """Set ray file generation without polarization data.

        Returns
        -------
        ansys.speos.core.sensor.SensorIrradiance
            Irradiance sensor.
        """
        self._sensor_instance.irradiance_properties.ray_file_type = (
            ProtoScene.SensorInstance.EnumRayFileType.RayFileClassic
        )
        return self

    def set_ray_file_type_polarization(self) -> SensorIrradiance:
        """Set ray file generation with the polarization data for each ray.

        Returns
        -------
        ansys.speos.core.sensor.SensorIrradiance
            Irradiance sensor.
        """
        self._sensor_instance.irradiance_properties.ray_file_type = (
            ProtoScene.SensorInstance.EnumRayFileType.RayFilePolarization
        )
        return self

    def set_ray_file_type_tm25(self) -> SensorIrradiance:
        """Set ray file generation: a .tm25ray file with polarization data for each ray.

        Returns
        -------
        ansys.speos.core.sensor.SensorIrradiance
            Irradiance sensor.
        """
        self._sensor_instance.irradiance_properties.ray_file_type = (
            ProtoScene.SensorInstance.EnumRayFileType.RayFileTM25
        )
        return self

    def set_ray_file_type_tm25_no_polarization(self) -> SensorIrradiance:
        """Set ray file generation: a .tm25ray file without polarization data.

        Returns
        -------
        ansys.speos.core.sensor.SensorIrradiance
            Irradiance sensor.
        """
        self._sensor_instance.irradiance_properties.ray_file_type = (
            ProtoScene.SensorInstance.EnumRayFileType.RayFileTM25NoPolarization
        )
        return self

    def set_layer_type_none(self) -> SensorIrradiance:
        """
        Define layer separation type as None.

        Returns
        -------
        ansys.speos.core.sensor.SensorIrradiance
            irradiance class instance

        """
        self._sensor_instance.irradiance_properties.layer_type_none.SetInParent()
        self._layer_type = None
        return self

    def set_layer_type_source(self) -> SensorIrradiance:
        """Define layer separation as by source.

        Returns
        -------
        ansys.speos.core.sensor.SensorIrradiance
           irradiance class instance

        """
        self._sensor_instance.irradiance_properties.layer_type_source.SetInParent()
        self._layer_type = None
        return self

    def set_layer_type_face(self) -> BaseSensor.LayerTypeFace:
        """Define layer separation as by face.

        Returns
        -------
        ansys.speos.core.sensor.BaseSensor.LayerTypeFace
            LayerTypeFace property instance
        """
        if self._layer_type is None and self._sensor_instance.irradiance_properties.HasField(
            "layer_type_face"
        ):
            # Happens in case of project created via load of speos file
            self._layer_type = BaseSensor.LayerTypeFace(
                layer_type_face=self._sensor_instance.irradiance_properties.layer_type_face,
                default_values=False,
                stable_ctr=True,
            )
        elif not isinstance(self._layer_type, BaseSensor.LayerTypeFace):
            # if the _layer_type is not LayerTypeFace then we create a new type.
            self._layer_type = BaseSensor.LayerTypeFace(
                layer_type_face=self._sensor_instance.irradiance_properties.layer_type_face,
                stable_ctr=True,
            )
        elif (
            self._layer_type._layer_type_face
            is not self._sensor_instance.irradiance_properties.layer_type_face
        ):
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._layer_type._layer_type_face = (
                self._sensor_instance.irradiance_properties.layer_type_face
            )
        return self._layer_type

    def set_layer_type_sequence(self) -> BaseSensor.LayerTypeSequence:
        """Define layer separation as by sequence.

        Returns
        -------
        ansys.speos.core.sensor.BaseSensor.LayerTypeSequence
            LayerTypeSequence property instance
        """
        if self._layer_type is None and self._sensor_instance.irradiance_properties.HasField(
            "layer_type_sequence"
        ):
            # Happens in case of project created via load of speos file
            self._layer_type = BaseSensor.LayerTypeSequence(
                layer_type_sequence=self._sensor_instance.irradiance_properties.layer_type_sequence,
                default_values=False,
                stable_ctr=True,
            )
        elif not isinstance(self._layer_type, BaseSensor.LayerTypeSequence):
            # if the _type is not LayerTypeSequence then we create a new type.
            self._layer_type = BaseSensor.LayerTypeSequence(
                layer_type_sequence=self._sensor_instance.irradiance_properties.layer_type_sequence,
                stable_ctr=True,
            )
        elif (
            self._layer_type._layer_type_sequence
            is not self._sensor_instance.irradiance_properties.layer_type_sequence
        ):
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._layer_type._layer_type_sequence = (
                self._sensor_instance.irradiance_properties.layer_type_sequence
            )
        return self._layer_type

    def set_layer_type_polarization(self) -> SensorIrradiance:
        """Define layer separation as by polarization.

        Returns
        -------
        ansys.speos.core.sensor.SensorIrradiance
            Irradiance class instance
        """
        self._sensor_instance.irradiance_properties.layer_type_polarization.SetInParent()
        self._layer_type = None
        return self

    def set_layer_type_incidence_angle(
        self,
    ) -> BaseSensor.LayerTypeIncidenceAngle:
        """Define layer separation as by incidence angle.

        Returns
        -------
        ansys.speos.core.sensor.BaseSensor.LayerTypeIncidenceAngle
            LayerTypeIncidenceAngle property instance
        """
        if self._layer_type is None and self._sensor_instance.irradiance_properties.HasField(
            "layer_type_incidence_angle"
        ):
            # Happens in case of project created via load of speos file
            self._layer_type = BaseSensor.LayerTypeIncidenceAngle(
                layer_type_incidence_angle=self._sensor_instance.irradiance_properties.layer_type_incidence_angle,
                default_values=False,
                stable_ctr=True,
            )
        elif not isinstance(self._layer_type, BaseSensor.LayerTypeIncidenceAngle):
            # if the _layer_type is not LayerTypeIncidenceAngle then we create a new type.
            self._layer_type = BaseSensor.LayerTypeIncidenceAngle(
                layer_type_incidence_angle=self._sensor_instance.irradiance_properties.layer_type_incidence_angle,
                stable_ctr=True,
            )
        elif (
            self._layer_type._layer_type_incidence_angle
            is not self._sensor_instance.irradiance_properties.layer_type_incidence_angle
        ):
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._layer_type._layer_type_incidence_angle = (
                self._sensor_instance.irradiance_properties.layer_type_incidence_angle
            )
        return self._layer_type

    def set_output_face_geometries(
        self, geometries: Optional[List[GeoRef]] = None
    ) -> SensorIrradiance:
        """Select output faces for inverse simulation optimization.

        Parameters
        ----------
        geometries : List[ansys.speos.core.geo_ref.GeoRef]
            List of geometries that will be considered as output faces.
            By default, ``[]``, ie no output faces.

        Returns
        -------
        ansys.speos.core.sensor.SensorIrradiance
            Irradiance sensor.
        """
        if not geometries:
            self._sensor_instance.irradiance_properties.ClearField("output_face_geometries")
        else:
            self._sensor_instance.irradiance_properties.output_face_geometries.geo_paths[:] = [
                gr.to_native_link() for gr in geometries
            ]
        return self


class SensorRadiance(BaseSensor):
    """Sensor feature: Radiance.

    By default, regarding inherent characteristics, a radiance sensor of type photometric is chosen.
    By default, regarding properties, an axis system is selected to position the sensor and no layer
    separation is chosen.

    Parameters
    ----------
    project : ansys.speos.core.project.Project
        Project that will own the feature.
    name : str
        Name of the feature.
    description : str
        Description of the feature.
        By default, ``""``.
    metadata : Optional[Mapping[str, str]]
        Metadata of the feature.
        By default, ``{}``.
    sensor_instance : ansys.api.speos.scene.v2.scene_pb2.Scene.SensorInstance, optional
        Sensor instance to provide if the feature does not has to be created from scratch
        By default, ``None``, means that the feature is created from scratch by default.
    default_values : bool
        Uses default values when True.
        By default, ``True``.
    """

    def __init__(
        self,
        project: project.Project,
        name: str,
        description: str = "",
        metadata: Optional[Mapping[str, str]] = None,
        sensor_instance: Optional[ProtoScene.SensorInstance] = None,
        default_values: bool = True,
    ) -> None:
        if metadata is None:
            metadata = {}

        super().__init__(
            project=project,
            name=name,
            description=description,
            metadata=metadata,
            sensor_instance=sensor_instance,
        )

        # Attribute gathering more complex radiance type
        self._type = None

        # Attribute gathering more complex layer type
        self._layer_type = None

        # Attribute to keep track of sensor dimensions object
        self._sensor_dimensions = self.Dimensions(
            sensor_dimensions=self._sensor_template.radiance_sensor_template.dimensions,
            default_values=default_values,
            stable_ctr=True,
        )

        if default_values:
            # Default values template
            self.set_focal().set_integration_angle().set_type_photometric()
            # Default values properties
            self.set_axis_system().set_layer_type_none()

    @property
    def visual_data(self) -> _VisualData:
        """Property containing radiance sensor visualization data.

        Returns
        -------
        _VisualData
            Instance of VisualData Class for pyvista.PolyData of feature faces, coordinate_systems.

        """
        if self._visual_data.updated:
            return self._visual_data
        else:
            feature_pos_info = self.get(key="axis_system")
            feature_radiance_pos = np.array(feature_pos_info[:3])
            feature_radiance_x_dir = np.array(feature_pos_info[3:6])
            feature_radiance_y_dir = np.array(feature_pos_info[6:9])
            feature_radiance_z_dir = np.array(feature_pos_info[9:12])
            feature_x_start = float(self.get(key="x_start"))
            feature_x_end = float(self.get(key="x_end"))
            feature_y_start = float(self.get(key="y_start"))
            feature_y_end = float(self.get(key="y_end"))
            feature_radiance_focal = float(self.get(key="focal"))

            # radiance sensor
            p1 = (
                feature_radiance_pos
                + feature_radiance_x_dir * feature_x_end
                + feature_radiance_y_dir * feature_y_end
            )
            p2 = (
                feature_radiance_pos
                + feature_radiance_x_dir * feature_x_end
                + feature_radiance_y_dir * feature_y_start
            )
            p3 = (
                feature_radiance_pos
                + feature_radiance_x_dir * feature_x_start
                + feature_radiance_y_dir * feature_y_start
            )
            p4 = (
                feature_radiance_pos
                + feature_radiance_x_dir * feature_x_start
                + feature_radiance_y_dir * feature_y_end
            )
            self._visual_data.add_data_rectangle([p1, p2, p3])

            p5 = feature_radiance_pos + feature_radiance_z_dir * feature_radiance_focal
            self._visual_data.add_data_triangle([p1, p2, p5])
            self._visual_data.add_data_triangle([p3, p4, p5])
            self._visual_data.add_data_triangle([p1, p4, p5])
            self._visual_data.add_data_triangle([p2, p3, p5])

            # radiance direction
            self._visual_data.coordinates.origin = feature_radiance_pos
            self._visual_data.coordinates.x_axis = feature_radiance_x_dir * max(
                feature_y_end - feature_y_start, feature_x_end - feature_x_start
            )
            self._visual_data.coordinates.y_axis = feature_radiance_y_dir * max(
                feature_y_end - feature_y_start, feature_x_end - feature_x_start
            )

            self._visual_data.updated = True
            return self._visual_data

    @property
    def dimensions(self) -> BaseSensor.Dimensions:
        """Property containing all options in regard to the Dimensions sensor properties.

        Returns
        -------
        ansys.speos.core.sensor.BaseSensor.Dimensions
            Instance of Dimensions Class for this sensor feature
        """
        return self._sensor_dimensions

    @property
    def type(self) -> str:
        """Type of sensor.

        Returns
        -------
        str
            Sensor type as string
        """
        if type(self._type) is str:
            return self._type
        elif isinstance(self._type, BaseSensor.Colorimetric):
            return "Colorimetric"
        elif isinstance(self._type, BaseSensor.Spectral):
            return "Spectral"
        else:
            return self._type

    @property
    def colorimetric(self) -> Union[None, BaseSensor.Colorimetric]:
        """Property containing all options in regard to the Colorimetric sensor properties.

        Returns
        -------
        ansys.speos.core.sensor.BaseSensor.Colorimetric
            Instance of Colorimetric Class for this sensor feature
        """
        if isinstance(self._type, BaseSensor.Colorimetric):
            return self._type
        else:
            return None

    @property
    def spectral(self) -> Union[None, BaseSensor.Spectral]:
        """Property containing all options in regard to the Spectral sensor properties.

        Returns
        -------
        ansys.speos.core.sensor.BaseSensor.Spectral
            Instance of Spectral Class for this sensor feature
        """
        if isinstance(self._type, BaseSensor.Spectral):
            return self._type
        else:
            return None

    @property
    def layer(
        self,
    ) -> Union[None, BaseSensor.LayerTypeFace, BaseSensor.LayerTypeSequence]:
        """Property containing all options in regard to the layer separation property.

        Returns
        -------
        Union[\
            None,\
            ansys.speos.core.sensor.BaseSensor.LayerTypeFace,\
            ansys.speos.core.sensor.BaseSensor.LayerTypeSequence\
        ]
            Instance of Layer type Class for this sensor feature
        """
        return self._layer_type

    def set_dimensions(self) -> BaseSensor.Dimensions:
        """Set the dimensions of the sensor.

        Returns
        -------
        ansys.speos.core.sensor.BaseSensor.Dimensions
            Dimension class
        """
        if (
            self._sensor_dimensions._sensor_dimensions
            is not self._sensor_template.radiance_sensor_template.dimensions
        ):
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._sensor_dimensions._sensor_dimensions = (
                self._sensor_template.radiance_sensor_template.dimensions
            )
        return self._sensor_dimensions

    def set_type_photometric(self) -> SensorRadiance:
        """Set type photometric.

        The sensor considers the visible spectrum and gets the results in lm/m2 or lx.

        Returns
        -------
        ansys.speos.core.sensor.SensorRadiance
            Radiance sensor.
        """
        self._sensor_template.radiance_sensor_template.sensor_type_photometric.SetInParent()
        self._type = None
        return self

    def set_type_colorimetric(self) -> BaseSensor.Colorimetric:
        """Set type colorimetric.

        The sensor will generate color results without any spectral data or layer separation
        in lx or W//m2.

        Returns
        -------
        ansys.speos.core.sensor.BaseSensor.Colorimetric
            Colorimetric type.
        """
        if self._type is None and self._sensor_template.radiance_sensor_template.HasField(
            "sensor_type_colorimetric"
        ):
            # Happens in case of project created via load of speos file
            self._type = BaseSensor.Colorimetric(
                sensor_type_colorimetric=self._sensor_template.radiance_sensor_template.sensor_type_colorimetric,
                default_values=False,
                stable_ctr=True,
            )
        elif not isinstance(self._type, BaseSensor.Colorimetric):
            # if the _type is not Colorimetric then we create a new type.
            self._type = BaseSensor.Colorimetric(
                sensor_type_colorimetric=self._sensor_template.radiance_sensor_template.sensor_type_colorimetric,
                stable_ctr=True,
            )
        elif (
            self._type._sensor_type_colorimetric
            is not self._sensor_template.radiance_sensor_template.sensor_type_colorimetric
        ):
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._type._sensor_type_colorimetric = (
                self._sensor_template.radiance_sensor_template.sensor_type_colorimetric
            )
        return self._type

    def set_type_radiometric(self) -> SensorRadiance:
        """Set type radiometric.

        The sensor considers the entire spectrum and gets the results in W/m2.

        Returns
        -------
        ansys.speos.core.sensor.SensorRadiance
            Radiance sensor.
        """
        self._sensor_template.radiance_sensor_template.sensor_type_radiometric.SetInParent()
        self._type = None
        return self

    def set_type_spectral(self) -> BaseSensor.Spectral:
        """Set type spectral.

        The sensor will generate color results and spectral data separated by wavelength
        in lx or W/m2.

        Returns
        -------
        ansys.speos.core.sensor.BaseSensor.Spectral
            Spectral type.
        """
        if self._type is None and self._sensor_template.radiance_sensor_template.HasField(
            "sensor_type_spectral"
        ):
            # Happens in case of project created via load of speos file
            self._type = BaseSensor.Spectral(
                sensor_type_spectral=self._sensor_template.radiance_sensor_template.sensor_type_spectral,
                default_values=False,
                stable_ctr=True,
            )
        elif not isinstance(self._type, BaseSensor.Spectral):
            # if the _type is not Spectral then we create a new type.
            self._type = BaseSensor.Spectral(
                sensor_type_spectral=self._sensor_template.radiance_sensor_template.sensor_type_spectral,
                stable_ctr=True,
            )
        elif (
            self._type._sensor_type_spectral
            is not self._sensor_template.radiance_sensor_template.sensor_type_spectral
        ):
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._type._sensor_type_spectral = (
                self._sensor_template.radiance_sensor_template.sensor_type_spectral
            )
        return self._type

    def set_focal(self, value: float = 250) -> SensorRadiance:
        """Set the focal value.

        Parameters
        ----------
        value : float
            Focal (mm).
            By default, ``250``.

        Returns
        -------
        ansys.speos.core.sensor.SensorRadiance
            Radiance sensor.
        """
        self._sensor_template.radiance_sensor_template.focal = value
        return self

    def set_integration_angle(self, value: float = 5) -> SensorRadiance:
        """Set the integration angle.

        Parameters
        ----------
        value : float
            integration angle (degree)
            By default, ``5``.

        Returns
        -------
        ansys.speos.core.sensor.SensorRadiance
            Radiance sensor.
        """
        self._sensor_template.radiance_sensor_template.integration_angle = value
        return self

    def set_axis_system(self, axis_system: Optional[List[float]] = None) -> SensorRadiance:
        """Set position of the sensor.

        Parameters
        ----------
        axis_system : Optional[List[float]]
            Position of the sensor [Ox Oy Oz Xx Xy Xz Yx Yy Yz Zx Zy Zz].
            By default, ``[0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1]``.

        Returns
        -------
        ansys.speos.core.sensor.SensorRadiance
            Radiance sensor.
        """
        if axis_system is None:
            axis_system = [0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1]
        self._sensor_instance.radiance_properties.axis_system[:] = axis_system
        return self

    def set_observer_point(self, value: Optional[List[float]] = None) -> SensorRadiance:
        """Set the position of the observer point.

        This is optional, because the focal length is used by default.
        Choosing to set an observer point will make the focal length ignored.

        Parameters
        ----------
        value : List[float], optional
            Position of the observer point [Ox Oy Oz].
            By default, ``None``. None means that the focal length is used.

        Returns
        -------
        ansys.speos.core.sensor.SensorRadiance
            Radiance sensor.
        """
        if not value:
            self._sensor_instance.radiance_properties.ClearField("observer_point")
        else:
            self._sensor_instance.radiance_properties.observer_point[:] = value
        return self

    def set_layer_type_none(self) -> SensorRadiance:
        """Define layer separation type as None.

        Returns
        -------
        ansys.speos.core.sensor.SensorRadiance
            Radiance sensor

        """
        self._sensor_instance.radiance_properties.layer_type_none.SetInParent()
        self._layer_type = None
        return self

    def set_layer_type_source(self) -> SensorRadiance:
        """Define layer separation as by source.

        Returns
        -------
        ansys.speos.core.sensor.SensorRadiance
            Radiance sensor

        """
        self._sensor_instance.radiance_properties.layer_type_source.SetInParent()
        self._layer_type = None
        return self

    def set_layer_type_face(self) -> BaseSensor.LayerTypeFace:
        """Define layer separation as by face.

        Returns
        -------
        ansys.speos.core.sensor.BaseSensor.LayerTypeFace
            LayerTypeFace property instance
        """
        if self._layer_type is None and self._sensor_instance.radiance_properties.HasField(
            "layer_type_face"
        ):
            # Happens in case of project created via load of speos file
            self._layer_type = BaseSensor.LayerTypeFace(
                layer_type_face=self._sensor_instance.radiance_properties.layer_type_face,
                default_values=False,
                stable_ctr=True,
            )
        elif not isinstance(self._layer_type, BaseSensor.LayerTypeFace):
            # if the _layer_type is not LayerTypeFace then we create a new type.
            self._layer_type = BaseSensor.LayerTypeFace(
                layer_type_face=self._sensor_instance.radiance_properties.layer_type_face,
                stable_ctr=True,
            )
        elif (
            self._layer_type._layer_type_face
            is not self._sensor_instance.radiance_properties.layer_type_face
        ):
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._layer_type._layer_type_face = (
                self._sensor_instance.radiance_properties.layer_type_face
            )
        return self._layer_type

    def set_layer_type_sequence(self) -> BaseSensor.LayerTypeSequence:
        """Define layer separation as by sequence.

        Returns
        -------
        ansys.speos.core.sensor.BaseSensor.LayerTypeSequence
            LayerTypeSequence property instance
        """
        if self._layer_type is None and self._sensor_instance.radiance_properties.HasField(
            "layer_type_sequence"
        ):
            # Happens in case of project created via load of speos file
            self._layer_type = BaseSensor.LayerTypeSequence(
                layer_type_sequence=self._sensor_instance.radiance_properties.layer_type_sequence,
                default_values=False,
                stable_ctr=True,
            )
        elif not isinstance(self._layer_type, BaseSensor.LayerTypeSequence):
            # if the _layer_type is not LayerTypeSequence then we create a new type.
            self._layer_type = BaseSensor.LayerTypeSequence(
                layer_type_sequence=self._sensor_instance.radiance_properties.layer_type_sequence,
                stable_ctr=True,
            )
        elif (
            self._layer_type._layer_type_sequence
            is not self._sensor_instance.radiance_properties.layer_type_sequence
        ):
            # Happens in case of feature reset (to be sure to always modify correct data)
            self._layer_type._layer_type_sequence = (
                self._sensor_instance.radiance_properties.layer_type_sequence
            )
        return self._layer_type
